import copy
import datetime
import shlex
import cmd
import readline
import os
import sys
import time

import sh
from typing import Any, Callable, List, Optional, Union, Iterable, cast
import shutil

import click
import peewee as pw
from click import style
from kong.model.job import color_dict

from .util import rjust, shorten_path, Spinner
from .state import DoesNotExist
from .config import APP_NAME, APP_DIR
from .logger import logger
from .model import Job, Folder
from . import state

history_file = os.path.join(APP_DIR, "history")


def complete_path(cwd: Folder, path: str) -> List[str]:
    logger.debug("Completion of '%s'", path)
    if path.endswith("/"):
        folder = Folder.find_by_path(cwd, path)
        prefix = ""
    else:
        head, prefix = os.path.split(path)
        folder = Folder.find_by_path(cwd, head)

    assert folder is not None
    options = []
    for child in folder.children:
        if child.name.startswith(prefix):
            options.append(child.name + "/")
    return options


def add_completion(*names: str) -> Callable[[type], type]:
    def decorator(cls: type) -> type:
        for name in names:
            method_name = f"complete_{name}"

            def handler(
                self: type, text: str, line: str, begidx: int, endidx: int
            ) -> List[str]:
                logger.debug(
                    "%s: text: %s, line: %s, begidx: %d, endidx: %d",
                    method_name,
                    text,
                    line,
                    begidx,
                    endidx,
                )

                # find component
                parts = shlex.split(line)
                base: Optional[str] = None
                for i, part in enumerate(parts):  # pragma: no branch
                    prelength = len(" ".join(parts[: i + 1]))
                    if prelength >= begidx:
                        base = part
                        break
                assert base is not None, "Error extracting active part"

                return complete_path(self.state.cwd, base)  # type: ignore

            setattr(cls, method_name, handler)
        return cls

    return decorator


def parse_arguments(fn: Any) -> Callable[[Any, str], None]:
    _, prog_name = fn.__name__.split("_", 1)

    orig_fn = copy.deepcopy(fn)
    fn = click.pass_obj(fn)
    command = click.command()(fn)

    def wrapped(self: Any, argstr: str) -> None:
        argv = shlex.split(argstr)
        logger.debug("%s", argv)
        try:
            command.main(
                args=argv,
                prog_name=prog_name,
                standalone_mode=False,
                obj=self,
                help_option_names=["-h", "--help"],
            )
        except click.MissingParameter:
            click.echo(f"usage: {fn.__doc__}")

    wrapped.__doc__ = fn.__doc__  # type: ignore
    wrapped.__name__ = fn.__name__  # type: ignore
    wrapped.__orig_fn__ = orig_fn  # type: ignore

    return wrapped


@add_completion(
    "ls",
    "mkdir",
    "mv",
    "rm",
    "cd",
    "submit_job",
    "kill_job",
    "info",
    "resubmit_job",
    "wait",
)
class Repl(cmd.Cmd):
    intro = f"This is {APP_NAME} shell"
    prompt = f"({APP_NAME} > /) "

    def __init__(self, state: state.State) -> None:
        self.state = state
        super().__init__()

    def precmd(self, line: str) -> str:
        if line != "":
            logger.debug("called '%s'", line)
        return line

    def onecmd(self, *args: str) -> bool:
        try:
            res = super().onecmd(*args)
            # logger.debug(
            #     "Writing history of length %d to file %s",
            #     self.state.config.history_length,
            #     history_file,
            # )
            readline.set_history_length(self.state.config.history_length)
            readline.write_history_file(history_file)
            return res
        except (BaseException, Exception) as e:
            logger.debug("Exception occured", exc_info=True)
            click.secho(f"{e}", fg="red")
        return False

    @parse_arguments
    @click.argument("dir", default="", required=False)
    @click.option("--refresh", "-r", is_flag=True)
    @click.option("--recursive", "-R", is_flag=True)
    @click.option("--json", "as_json", is_flag=True)
    def do_ls(self, dir: str, refresh: bool, recursive: bool, as_json: bool) -> None:
        "List the current directory content"
        try:
            width, height = shutil.get_terminal_size((80, 40))

            with Spinner("Getting info", persist=False, enabled=refresh or recursive):
                folders, jobs = self.state.ls(dir, refresh=refresh)

                if recursive:
                    arg_folder = Folder.find_by_path(self.state.cwd, dir)
                    assert arg_folder is not None  # should be a folder
                    jobs = arg_folder.jobs_recursive()

                if refresh:
                    jobs = cast(list, self.state.refresh_jobs(jobs))

            if len(folders) > 0:
                folder_name_length = max([len(f.name) for f in folders])

                headers = ("name", "job counts")

                folder_name_length = max(folder_name_length, len(headers[0]))

                click.echo(
                    headers[0].ljust(folder_name_length)
                    + " "
                    + headers[1].rjust(width - folder_name_length - 1)
                )

                click.echo(
                    "-" * folder_name_length
                    + " "
                    + "-" * (width - folder_name_length - 1)
                )

                for folder in folders:
                    folder_jobs = folder.jobs_recursive()
                    # accumulate counts
                    # @TODO: SLOW! Optimize to query
                    counts = {k: 0 for k in Job.Status}

                    for job in folder_jobs:
                        counts[job.status] += 1

                    output = ""
                    for k, c in counts.items():
                        output += style(f" {c:> 6d}{k.name[:1]}", fg=color_dict[k])
                    output = folder.name.ljust(folder_name_length) + rjust(
                        output, width - folder_name_length
                    )

                    click.echo(output)

                click.echo("")

            if len(jobs) > 0:
                headers_jobs = (
                    "job id",
                    "batch job id",
                    "created",
                    "updated",
                    "status",
                )

                name_length = max(
                    max([len(str(j.job_id)) for j in jobs]), len(headers_jobs[0])
                )

                status_len = len("SUBMITTED")
                status_len = max(status_len, len(headers_jobs[-1]))

                def dtfmt(dt: datetime.datetime) -> str:
                    return dt.strftime("%Y-%m-%d %H:%M:%S")

                datetime_len = len(dtfmt(jobs[0].updated_at))

                bjobid_len = (
                    width
                    - name_length
                    - status_len
                    - 2 * datetime_len
                    - len(headers_jobs)
                )
                bjobid_len = max(bjobid_len, len(headers_jobs[1]))

                click.echo(
                    headers_jobs[0].rjust(name_length)
                    + " "
                    + headers_jobs[1].rjust(bjobid_len)
                    + " "
                    + headers_jobs[2].ljust(datetime_len)
                    + " "
                    + headers_jobs[3].ljust(datetime_len)
                    + " "
                    + headers_jobs[4].ljust(status_len)
                )
                click.echo(
                    "-" * name_length
                    + " "
                    + "-" * bjobid_len
                    + " "
                    + "-" * datetime_len
                    + " "
                    + "-" * datetime_len
                    + " "
                    + "-" * status_len
                )

                for job in jobs:
                    job_id = str(job.job_id).rjust(name_length)
                    if job.batch_job_id is None:
                        batch_job_id = " " * bjobid_len
                    else:
                        batch_job_id = job.batch_job_id.rjust(bjobid_len)
                    _, status_name = str(job.status).split(".", 1)
                    color = color_dict[job.status]

                    # table.append_row(())

                    click.secho(
                        f"{job_id} {batch_job_id} {dtfmt(job.created_at)} {dtfmt(job.updated_at)} {status_name}",
                        fg=color,
                    )

        except pw.DoesNotExist:
            click.secho(f"Folder {dir} does not exist", fg="red")

    @parse_arguments
    @click.argument("path")
    @click.option("--create-parent", "-p", is_flag=True)
    def do_mkdir(self, path: str, create_parent: bool) -> None:
        "Create a directory at the current location"
        try:
            self.state.mkdir(path, create_parent=create_parent)
        except state.CannotCreateError:
            click.secho(f"Cannot create folder at '{path}'", fg="red")
        except pw.IntegrityError:
            click.secho(
                f"Folder {path} in {self.state.cwd.path} already exists", fg="red"
            )

    @parse_arguments
    @click.argument("name", required=False, default="")
    def do_cd(self, name: str = "") -> None:
        # find the folder
        try:
            self.state.cd(name)
        except pw.DoesNotExist:
            click.secho(f"Folder {name} does not exist", fg="red")
        self.prompt = f"({APP_NAME} > {shorten_path(self.state.cwd.path, 40)}) "

    @parse_arguments
    @click.argument("src")
    @click.argument("dest")
    def do_mv(self, src: str, dest: str) -> None:
        items: List[Union[Job, Folder]] = self.state.mv(src, dest)
        names = []
        for item in items:
            if isinstance(item, Job):
                names.append(str(item.job_id))
            else:
                names.append(item.name)

        click.secho(f"Moved {', '.join(names)} -> {dest}")

    @parse_arguments
    @click.argument("job_arg")
    @click.option("--refresh", "-r", is_flag=True)
    @click.option("--recursive", "-R", is_flag=True)
    @click.option("--full", is_flag=True)
    def do_info(self, job_arg: str, refresh: bool, recursive: bool, full: bool) -> None:
        jobs = self.state.get_jobs(job_arg, recursive)
        if refresh:
            jobs = list(self.state.refresh_jobs(jobs))

        for job in jobs:
            click.echo(job)
            for field in (
                "batch_job_id",
                "driver",
                "folder",
                "command",
                "cores",
                "status",
                "created_at",
                "updated_at",
            ):
                fg: Optional[str] = None
                if field == "status":
                    fg = color_dict[job.status]
                if field == "command" and not full:
                    cmd = job.command
                    if len(cmd) > 500:
                        cmd = cmd[:500] + "..."
                    click.secho(f"{field}: {cmd}", fg=fg)
                    continue

                click.secho(f"{field}: {str(getattr(job, field))}", fg=fg)
            click.echo("data:")
            for k, v in job.data.items():
                click.secho(f"- {k}: {v}")

    @parse_arguments
    @click.argument("job")
    @click.option("--recursive", "-R", is_flag=True)
    def do_rm(self, job: str, recursive: bool) -> None:
        try:
            if self.state.rm(
                job, recursive=recursive, confirm=lambda s: click.confirm(s)
            ):
                click.echo(f"{job} is gone")
        except state.CannotRemoveRoot:
            click.secho("Cannot delete root folder", fg="red")
        except DoesNotExist:
            click.secho(f"Folder {job} does not exist", fg="red")

    @parse_arguments
    def do_cwd(self) -> None:
        "Show the current location"
        click.echo(self.state.cwd.path)

    @parse_arguments
    @click.argument("command", nargs=-1)
    @click.option("--cores", "-c", type=int, default=1)
    def do_create_job(self, command: List[str], cores: int) -> None:
        if len(command) == 0:
            click.secho("Please provide a command to run", fg="red")
            return
        command_str = " ".join(command)

        job = self.state.create_job(command=command_str, cores=cores)
        click.secho(f"Created job {job}")

    @parse_arguments
    @click.argument("job")
    @click.option("--recursive", "-R", is_flag=True)
    def do_submit_job(self, job: str, recursive: bool) -> None:
        self.state.submit_job(job, click.confirm, recursive=recursive)

    @parse_arguments
    @click.argument("job")
    @click.option("--recursive", "-R", is_flag=True)
    def do_kill_job(self, job: str, recursive: bool) -> None:
        self.state.kill_job(job, recursive=recursive, confirm=click.confirm)

    @parse_arguments
    @click.argument("job")
    @click.option("--recursive", "-R", is_flag=True)
    @click.option("--failed", "-F", is_flag=True)
    def do_resubmit_job(self, job: str, recursive: bool, failed: bool) -> None:
        self.state.resubmit_job(
            job, click.confirm, recursive=recursive, failed_only=failed
        )

    @parse_arguments
    @click.argument("job_arg")
    @click.option("--refresh", "-r", is_flag=True)
    def do_status(self, job_arg: str, refresh: bool) -> None:
        """Print status of JOB"""
        jobs = self.state.get_jobs(job_arg)

        if refresh:
            self.state.refresh_jobs(jobs)

        for job in jobs:
            click.echo(f"{job}")

    @parse_arguments
    @click.argument("job")
    def do_update(self, job: str) -> None:
        jobs = self.state.get_jobs(job)
        self.state.refresh_jobs(jobs)

    @parse_arguments
    @click.argument("job_str")
    @click.option("--number-of-lines", "-n", default=20, type=int)
    def do_tail(self, job_str: str, number_of_lines: int) -> None:
        from sh import tail  # type: ignore

        jobs = self.state.get_jobs(job_str)
        assert len(jobs) == 1
        job = jobs[0]

        if not os.path.exists(job.data["stdout"]):
            with Spinner(
                text=f"Waiting for job to create stdout file '{job.data['stdout']}'"
            ):
                while not os.path.exists(job.data["stdout"]):
                    time.sleep(1)
                logger.info("Outfile exists now")

        width, _ = shutil.get_terminal_size((80, 40))
        hw = width // 2
        click.echo("=" * hw + " STDOUT " + "=" * (width - hw - 8))

        def show(line: str) -> None:  # pragma: no cover
            sys.stdout.write(line)

        try:
            proc = tail(
                "-n",
                number_of_lines,
                "-f",
                job.data["stdout"],
                _bg=True,
                _bg_exc=False,
                _out=show,
            )
            try:
                proc.wait()
            except KeyboardInterrupt:  # pragma: no cover
                proc.terminate()
                proc.wait()

        except sh.SignalException_SIGTERM:  # pragma: no cover
            pass

    @parse_arguments
    @click.argument("job_str")
    def do_less(self, job_str: str) -> None:
        jobs = self.state.get_jobs(job_str)
        assert len(jobs) == 1
        job = jobs[0]

        def reader() -> Iterable[str]:
            with open(job.data["stdout"]) as fp:
                line = fp.readline()
                yield line
                while line:
                    line = fp.readline()
                    yield line

        click.echo_via_pager(reader())

    @parse_arguments
    @click.argument("job_arg")
    @click.option("--notify/--no-notify", default=True)
    @click.option("--recursive", "-R", is_flag=True)
    @click.option("--poll-interval", "-i", type=int, default=None)
    def do_wait(
        self, job_arg: str, notify: bool, recursive: bool, poll_interval: int
    ) -> None:
        self.state.wait(
            job_arg, notify=notify, recursive=recursive, poll_interval=poll_interval
        )

    def do_exit(self, arg: str) -> bool:
        return True

    def do_EOF(self, arg: str) -> bool:
        return self.do_exit(arg)

    def preloop(self) -> None:
        if os.path.exists(history_file):
            # logger.debug("Loading history from %s", history_file)
            readline.read_history_file(history_file)
        else:
            logger.debug("No history file found")

    def postloop(self) -> None:
        pass

    def cmdloop(self, intro: Optional[Any] = None) -> Any:
        print(self.intro)
        while True:
            try:
                super().cmdloop(intro="")
                break
            except KeyboardInterrupt:
                print("^C")

    def emptyline(self) -> bool:
        # do nothing when called with empty
        pass



from unittest.mock import Mock

import pytest
import inspect

from kong.drivers import DriverMismatch
from kong.drivers.driver_base import DriverBase, checked_job
from kong.model import Job


@pytest.fixture
def noabc(monkeypatch):
    monkeypatch.setattr(
        "kong.drivers.driver_base.DriverBase.__abstractmethods__", set()
    )


@pytest.fixture
def driver(noabc, state):
    return DriverBase(state.config)


def test_driver_base_all_methods(driver):
    methods = inspect.getmembers(driver, predicate=inspect.ismethod)

    excl = ["__init__", "_check_driver"]

    for name, method in methods:
        if not hasattr(method, "__isabstractmethod__"):
            continue

        if not method.__isabstractmethod__:
            continue

        sig = inspect.signature(method)

        args = len(sig.parameters) * [None]

        try:
            method(*args)
        except NotImplementedError:
            pass

def test_init(noabc, state):
    DriverBase(None)
    DriverBase(state.config)


def test_check_driver(driver):
    job = Mock()

    class NotADriver:
        pass

    job.driver = NotADriver
    with pytest.raises(DriverMismatch):
        driver._check_driver(job)

    job.driver = DriverBase
    driver._check_driver(job)


def test_checked_job(driver):

    inner = Mock()
    wrapped = checked_job(inner)

    job = Mock()

    class NotADriver:
        pass

    job.driver = NotADriver
    with pytest.raises(DriverMismatch):
        wrapped(driver, job)
    assert inner.call_count == 0

    inner.reset_mock()

    job.driver = DriverBase
    wrapped(driver, job)
    inner.assert_called_once_with(driver, job)


def test_wait(driver, monkeypatch):
    wait_gen = Mock(return_value=iter([]))
    monkeypatch.setattr(driver, "wait_gen", wait_gen)

    driver.wait("*", progress=True)
    assert wait_gen.call_count == 1

    wait_gen.reset_mock()

    driver.wait("*", progress=False)
    assert wait_gen.call_count == 1

def test_make_log_path(driver, monkeypatch):
        monkeypatch.setitem(driver.config.data, "jobdir", "JOB_BASE")

        job = Mock()

        job.job_id = 123456
        assert driver.make_log_path(job).endswith("JOB_BASE/12/34/123456")

        job.job_id = 994784367
        assert driver.make_log_path(job).endswith("JOB_BASE/99/47/994784367")

        job.job_id = 456
        assert driver.make_log_path(job).endswith("JOB_BASE/00/04/000456")

        job.job_id = 3854
        assert driver.make_log_path(job).endswith("JOB_BASE/00/38/003854")


def test_make_output_path(driver, monkeypatch):
    monkeypatch.setitem(driver.config.data, "joboutputdir", "JOB_BASE")

    job = Mock()

    job.job_id = 123456
    assert driver.make_output_path(job).endswith("JOB_BASE/12/34/123456")

    job.job_id = 994784367
    assert driver.make_output_path(job).endswith("JOB_BASE/99/47/994784367")

    job.job_id = 456
    assert driver.make_output_path(job).endswith("JOB_BASE/00/04/000456")

    job.job_id = 3854
    assert driver.make_output_path(job).endswith("JOB_BASE/00/38/003854")



import datetime
from fnmatch import fnmatch
import os
import re
from typing import (
    List,
    Callable,
    Any,
    Union,
    Optional,
    Tuple,
    Sequence,
    Iterable,
    Iterator,
    cast,
)

import peewee as pw
from contextlib import contextmanager

from click import style
from kong.model.job import color_dict

from .util import Progress, Spinner, exhaust
from .drivers import DriverMismatch, get_driver
from .drivers.driver_base import DriverBase
from . import config
from .db import database
from . import model
from .model import Folder, Job
from .logger import logger


class CannotCreateError(RuntimeError):
    pass


class CannotRemoveRoot(RuntimeError):
    pass


class DoesNotExist(RuntimeError):
    pass


class CannotRemoveIsFolder(RuntimeError):
    pass


JobSpec = Union[str, int, Job]

Confirmation = Callable[[str], bool]


def YES(_: str) -> bool:
    return True


class State:
    def __init__(self, config: config.Config, cwd: Folder) -> None:
        self.config = config
        self.cwd = cwd
        self.default_driver: DriverBase = get_driver(self.config.default_driver)(
            self.config
        )

    @contextmanager  # type: ignore
    def pushd(self, folder: Union["Folder", str]) -> Iterator[None]:
        prev = self.cwd

        if isinstance(folder, Folder):
            self.cwd = folder
        elif isinstance(folder, str):
            _folder = Folder.find_by_path(self.cwd, folder)
            assert _folder is not None
            self.cwd = _folder
        else:
            raise TypeError("Argument is neither Folder nor str")

        try:
            yield
        finally:
            self.cwd = prev

    @classmethod
    def get_instance(cls) -> "State":
        cfg = config.Config()
        logger.debug("Initialized config: %s", cfg.data)

        logger.debug(
            "Initializing database '%s' at '%s'", config.APP_NAME, config.DB_FILE
        )
        database.init(config.DB_FILE)

        # ensure database is set up
        database.connect()
        database.create_tables([getattr(model, m) for m in model.__all__])

        cwd = Folder.get_root()

        return cls(cfg, cwd)

    def refresh_jobs(self, jobs: List[Job]) -> Sequence[Job]:
        logger.debug("Refreshing %d jobs", len(jobs))
        first_job: Job = jobs[0]
        # try bulk refresh first
        first_job.ensure_driver_instance(self.config)
        driver: DriverBase = first_job.driver_instance
        try:
            logger.debug("Attempting bulk mode sync using %s", driver.__class__)
            jobs = list(driver.bulk_sync_status(jobs))
        except DriverMismatch:
            # fall back to slow mode
            logger.debug("Bulk mode sync failed, falling back to slow loop mode")
            for job in jobs:
                job.ensure_driver_instance(self.config)
                job.get_status()
        return jobs

    def ls(
        self, path: str = ".", refresh: bool = False, recursive: bool = False
    ) -> Tuple[List["Folder"], List["Job"]]:
        "List the current directory content"
        logger.debug("%s", list(self.cwd.children))
        folder = Folder.find_by_path(self.cwd, path)
        if folder is None:
            raise pw.DoesNotExist()

        jobs = folder.jobs

        if refresh and len(jobs) > 0:
            if recursive:
                jobs = list(self.refresh_jobs(folder.jobs_recursive()))
            else:
                jobs = list(self.refresh_jobs(jobs))

        return list(folder.children), jobs

    def cd(self, target: Union[str, Folder] = ".") -> None:
        if isinstance(target, str):
            if target == "":
                folder = Folder.get_root()
            else:
                _folder = Folder.find_by_path(self.cwd, target)
                if _folder is None:
                    raise pw.DoesNotExist()
                folder = _folder
            self.cwd = folder
        elif isinstance(target, Folder):
            self.cwd = target
        else:
            raise TypeError(f"{target} is not str or Folder")

    def _mv_folder(self, source: Folder, dest: Union[str, Folder]) -> None:
        if isinstance(dest, Folder):
            # requested action: move source INTO dest
            source.parent = dest
            source.save()
        elif isinstance(dest, str):
            dest_folder = Folder.find_by_path(self.cwd, dest)
            if dest_folder is not None:
                # dest exists!
                # requested action: move source INTO dest
                source.parent = dest_folder
                source.save()
            else:
                # dest does not exist
                # requested action: RENAME source to dest and potentially move
                # head is new parent folder, tail is new name
                head, tail = os.path.split(dest)
                source.name = tail
                dest_folder = Folder.find_by_path(self.cwd, head)
                if dest_folder is None:
                    raise ValueError(f"Target folder {head} does not exist")

                source.parent = dest_folder
                source.save()
        else:
            raise TypeError(f"{dest} is neither string nor Folder")

    def _mv_folders(self, folders: List[Folder], dest: Union[str, Folder]) -> None:
        dest_folder: Folder
        if isinstance(dest, Folder):
            dest_folder = dest
        elif isinstance(dest, str):
            folder = Folder.find_by_path(self.cwd, dest)
            assert folder is not None
            dest_folder = folder
        else:
            raise TypeError(f"{dest} is neither string nor Folder")

        with database.atomic():
            Folder.update(parent=dest_folder, updated_at=datetime.datetime.now()).where(
                Folder.folder_id  # type: ignore
                << [f.folder_id for f in folders if f != dest_folder]  # type: ignore
            ).execute()

    def _mv_jobs(self, jobs: List[Job], dest: Union[str, Folder]) -> None:
        dest_folder: Optional[Folder]
        if isinstance(dest, Folder):
            dest_folder = dest
        elif isinstance(dest, str):
            dest_folder = Folder.find_by_path(self.cwd, dest)
            if dest_folder is None:
                raise ValueError(f"{dest} does not exists, and jobs cannot be renamed")
        else:
            raise TypeError(f"{dest} is neither string nor Folder")

        assert dest_folder is not None

        with database.atomic():
            Job.update(folder=dest_folder, updated_at=datetime.datetime.now()).where(
                Job.job_id << [j.job_id for j in jobs]  # type: ignore
            ).execute()

    def mv(
        self, source: Union[str, Job, Folder], dest: Union[str, Folder]
    ) -> List[Union[Job, Folder]]:
        # source might be: a job or a folder
        if isinstance(source, Folder):
            self._mv_folder(source, dest)
            return [source]
        elif isinstance(source, str):
            source_folder = Folder.find_by_path(self.cwd, source)
            if source_folder is not None:
                self._mv_folder(source_folder, dest)
                return [source_folder]
            else:
                # either job(s) or possibly a list of folders (*)
                jobs: List[Job] = []
                folders: List[Folder] = []

                try:
                    jobs = self.get_jobs(source)
                except ValueError:
                    pass

                try:
                    folders = self.get_folders(source)
                except ValueError:
                    pass

                if len(folders) == 0 and len(jobs) == 0:
                    raise DoesNotExist(f"No such folder or job: {source}")

                self._mv_jobs(jobs, dest)
                self._mv_folders(folders, dest)

                return list(folders) + list(jobs)  # type: ignore

        elif isinstance(source, Job):
            # is a job
            self._mv_jobs([source], dest)
            return [source]
        else:
            raise TypeError(f"{source} is not a Folder or a Job")

    def mkdir(
        self, path: str, exist_ok: bool = False, create_parent: bool = False
    ) -> Optional[Folder]:
        logger.debug("mkdir %s", path)
        found_folder = Folder.find_by_path(self.cwd, path)
        if found_folder is not None:
            if not exist_ok:
                raise CannotCreateError(f"Cannot create folder at {path}")
            else:
                return found_folder

        location: Optional[Folder] = None

        if path.startswith("/"):
            path = path[1:]

        head, tail = os.path.split(path)

        if create_parent and head != "":

            def create(p: str) -> Folder:
                head, tail = os.path.split(p)
                loc = Folder.find_by_path(self.cwd, head)
                if loc is None:
                    loc = create(head)
                subf = loc.subfolder(tail)
                if subf is not None:
                    return subf
                return Folder.create(name=tail, parent=loc)

            location = create(head)
        else:
            location = Folder.find_by_path(self.cwd, head)

        if location is None:
            raise CannotCreateError(f"Cannot create folder at '{path}'")

        logger.debug("Attempt to create folder named '%s' in '%s'", tail, location.path)

        return Folder.create(name=tail, parent=location)

    def rm(
        self,
        name: Union[str, Job, Folder],
        recursive: bool = False,
        confirm: Confirmation = lambda _: True,
    ) -> bool:
        jobs: List[Job] = []
        if isinstance(name, str):
            # string name, could be both
            if name == "/":
                raise CannotRemoveRoot()

            folders: List[Folder] = []

            try:
                folders = self.get_folders(name)
                if len(folders) > 0 and not recursive:
                    raise CannotRemoveIsFolder(
                        f"{name} matches {len(folders)} folder(s). Use recursive to delete"
                    )
                for folder in folders:
                    jobs += folder.jobs_recursive()
            except ValueError:
                pass
            try:
                jobs += self.get_jobs(name)
            except ValueError:
                pass

            if len(folders) == 0 and len(jobs) == 0:
                raise DoesNotExist(f"No such folder or job: {name}")

            if confirm(f"Remove {len(folders)} folders and {len(jobs)} jobs?"):

                if len(jobs) > 0:
                    first_job = jobs[0]
                    first_job.ensure_driver_instance(self.config)
                    driver = first_job.driver_instance

                    driver.bulk_remove(jobs)

                for folder in folders:
                    folder.delete_instance(recursive=True, delete_nullable=True)

                return True

            return False
        elif isinstance(name, Job):
            job = name

            if confirm(f"Delete job {job}?"):
                # need driver instance
                job.ensure_driver_instance(self.config)
                job.driver_instance.remove(job)
                return True
            return False
        elif isinstance(name, Folder):
            folder = name
            # jobs: List[Job] = []
            if not recursive:
                raise CannotRemoveIsFolder(f"Cannot remove {folder} non-recursively")
            jobs = folder.jobs_recursive()
            if confirm(f"Delete folder {folder.path} and {len(jobs)} jobs?"):
                if len(jobs) > 0:
                    first_job = jobs[0]
                    first_job.ensure_driver_instance(self.config)
                    driver = first_job.driver_instance

                    driver.bulk_remove(jobs)

                folder.delete_instance(recursive=True, delete_nullable=True)
                return True
            return False
        else:
            raise TypeError("Invalid rm target type given")

    def create_job(self, *args: Any, **kwargs: Any) -> Job:
        if "folder" in kwargs:
            raise ValueError("To submit to explicit folder, use driver directly")
        if "driver" in kwargs:
            raise ValueError("To submit with explicit driver, use it directly")
        kwargs["folder"] = self.cwd
        return self.default_driver.create_job(*args, **kwargs)

    def _extract_jobs(self, name: JobSpec, recursive: bool = False) -> List[Job]:
        jobs: List[Job] = []

        if isinstance(name, int):
            j = Job.get_or_none(name)
            if j is None:
                raise DoesNotExist(f"Did not find job with id {name}")
            jobs = [j]

        elif isinstance(name, str):
            # check if we have path component
            if name.isdigit():
                j = Job.get_or_none(int(name))
                if j is None:
                    raise DoesNotExist(f"Did not find job with path {name}")
                jobs = [j]
            else:
                head, tail = os.path.split(name)
                logger.debug("Getting job: head: %s, tail: %s", head, tail)

                r_m = re.match(r"(\d+)\.\.(\d+)", tail)

                if tail.isdigit():
                    # single job id, just get that
                    j = Job.get_or_none(int(tail))
                    if j is None:
                        raise DoesNotExist(f"Did not find job with path {name}")
                    jobs = [j]
                elif "*" in tail:
                    # "glob" jobs: get folder, and select all jobs
                    # this is half-recursive right now
                    # folder = Folder.find_by_path(self.cwd, head)
                    folders = self.get_folders(head)
                    jobs = sum([list(f.jobs) for f in folders], [])
                elif r_m is not None:
                    start = int(r_m.group(1))
                    end = int(r_m.group(2))

                    if start > end:
                        raise ValueError(f"Illegal job range: {tail}")

                    folder = Folder.find_by_path(self.cwd, head)
                    assert folder is not None

                    jobs = []
                    for job in folder.jobs:
                        if job.job_id < start or job.job_id > end:
                            continue
                        jobs.append(job)
                    return jobs
                else:
                    folder = Folder.find_by_path(self.cwd, name)
                    if not recursive or folder is None:
                        raise ValueError(f"{name} jobspec is not understood")

                    jobs = folder.jobs_recursive()
        elif isinstance(name, Job):
            jobs = [name]
        else:
            raise TypeError("Name is neither job id, path to job(s) nor job instance")

        return jobs

    def submit_job(
        self, name: JobSpec, confirm: Confirmation = YES, recursive: bool = False
    ) -> None:

        jobs: List[Job]
        if recursive and isinstance(name, str):
            # get folders, extract jobs from those
            folders = self.get_folders(name)
            logger.debug("Recursive, found %s folders", len(folders))
            jobs = sum([f.jobs_recursive() for f in folders], [])
        else:
            jobs = self._extract_jobs(name)

        if not confirm(f"Submit {len(jobs)} jobs?"):
            return

        assert len(jobs) > 0
        first_job = jobs[0]
        first_job.ensure_driver_instance(self.config)
        driver = first_job.driver_instance

        def job_iter() -> Iterable[Job]:
            job: Job
            for job in Progress(jobs, desc="Submitting jobs"):
                yield job

        driver.bulk_submit(job_iter())

    def kill_job(
        self, name: JobSpec, recursive: bool = False, confirm: Confirmation = YES
    ) -> None:
        jobs: List[Job]
        if recursive and isinstance(name, str):
            # get folders, extract jobs from thos
            folders = self.get_folders(name)
            logger.debug("Recursive, found %s folders", len(folders))
            jobs = sum([f.jobs_recursive() for f in folders], [])
        else:
            jobs = self._extract_jobs(name)

        if not confirm(f"Kill {len(jobs)}?"):
            return
        job: Job
        for job in Progress(jobs, desc="Killing jobs"):
            job.ensure_driver_instance(self.config)  # type: ignore
            job.kill()  # type: ignore

    def resubmit_job(
        self,
        name: JobSpec,
        confirm: Confirmation = YES,
        recursive: bool = False,
        failed_only: bool = False,
    ) -> None:
        jobs: List[Job]
        jobs = self._extract_jobs(name, recursive=recursive)

        if failed_only:
            jobs = [job for job in jobs if job.status == Job.Status.FAILED]

        if not confirm(f"Resubmit {len(jobs)} jobs?"):
            return

        assert len(jobs) > 0
        first_job = jobs[0]
        first_job.ensure_driver_instance(self.config)
        driver = first_job.driver_instance

        with Spinner(f"Preparing for resubmission for {len(jobs)} jobs"):
            jobs = list(driver.bulk_resubmit(jobs, do_submit=False))

        def job_iter() -> Iterable[Job]:
            job: Job
            for job in Progress(jobs, desc="Submitting jobs"):
                yield job

        driver.bulk_submit(job_iter())

    def get_jobs(self, name: JobSpec, recursive: bool = False) -> List[Job]:
        return self._extract_jobs(name, recursive)

    def get_folders(self, pattern: str) -> List[Folder]:
        head, tail = os.path.split(pattern)
        if "*" in tail:
            logger.debug("Pattern: %s, will glob", tail)
            folder: Optional[Folder] = None
            if head == "":
                folder = self.cwd
            else:
                folder = Folder.find_by_path(self.cwd, head)

            if folder is None:
                raise ValueError(f"No folder {head} found")

            if tail == "*":  # no need to match, just get all
                return folder.children
            else:
                folders = [f for f in folder.children if fnmatch(f.name, tail)]
                return folders
        else:
            folder = Folder.find_by_path(self.cwd, pattern)
            if folder is None:
                raise ValueError(f"No folder {pattern} found")
            return [folder]

    def wait(
        self, *args: Any, progress: bool = False, **kwargs: Any
    ) -> Optional[Iterable[List[Job]]]:
        it = self.wait_gen(*args, **kwargs)
        if progress:
            return it
        else:
            exhaust(it)
            return None

    def wait_gen(
        self,
        jobspec: JobSpec,
        recursive: bool = False,
        notify: bool = True,
        timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
    ) -> Iterable[List[Job]]:
        jobs: List[Job]
        jobs = self._extract_jobs(jobspec, recursive=recursive)
        assert len(jobs) > 0
        first_job = jobs[0]
        first_job.ensure_driver_instance(self.config)
        driver = first_job.driver_instance
        orig_jobs = jobs[:]

        try:
            with Spinner(text=f"Waiting for {len(jobs)} jobs") as s:
                for cur_jobs in cast(
                    Iterable[List[Job]],
                    driver.wait(
                        jobs,
                        timeout=timeout,
                        poll_interval=poll_interval,
                        progress=True,
                    ),
                ):
                    counts = {k: 0 for k in Job.Status}
                    for job in cur_jobs:
                        counts[job.status] += 1

                    out = [
                        style(f"{k.name[:1]}{v}", fg=color_dict[k])
                        for k, v in counts.items()
                    ]
                    s.text = f"Waiting for {len(jobs)} jobs: {', '.join(out)}"
                    yield cur_jobs

            orig_jobs = list(driver.bulk_sync_status(orig_jobs))
            counts = {k: 0 for k in Job.Status}
            for job in orig_jobs:
                counts[job.status] += 1

            out = [f"{k.name[:1]}{v}" for k, v in counts.items()]

            if notify:
                self.config.notifications.notify(
                    title="kong: Job wait complete",
                    message=f"Successfully waited for {len(jobs)} job(s) to finish:\n{', '.join(out)}",
                )
        except TimeoutError:
            if notify:
                self.config.notifications.notify(
                    title="kong: Job wait timeout",
                    message=f"Timeout waiting for {len(jobs)} job(s) after {timeout}s",
                )



import os
from datetime import timedelta

import pytimeparse
import typing
import socket
import getpass

import click
import yaml
from schema import And, Optional, Schema, Use
import notifiers  # type: ignore


APP_NAME = "kong"
APP_DIR = click.get_app_dir(APP_NAME, force_posix=True)
CONFIG_FILE = os.path.join(APP_DIR, "config.yml")
DB_FILE = os.path.join(APP_DIR, "database.sqlite")

slurm_schema = Schema(
    {
        Optional("account", default="account"): And(str, len),
        Optional("node_size", default=1): And(int, lambda i: i > 0),
        Optional("default_queue", default="queue"): And(str, len),
        Optional("sacct_delta", default=timedelta(weeks=4)): Use(
            lambda s: timedelta(seconds=pytimeparse.parse(s))
        ),
    }
)

htcondor_schema = Schema(
    {
        Optional("user", default=getpass.getuser()): And(str, len),
        Optional("default_universe", default="vanilla"): And(str, len),
        Optional("submitfile_extra", default=""): str,
    }
)

config_schema = Schema(
    {
        Optional(
            "default_driver", default="kong.drivers.local_driver.LocalDriver"
        ): And(str, len),
        Optional(
            "jobdir", default=lambda: os.path.join(APP_DIR, "jobdir")
        ): os.path.exists,
        Optional(
            "joboutputdir", default=lambda: os.path.join(APP_DIR, "joboutput")
        ): os.path.exists,
        Optional("history_length", default=1000): int,
        Optional("slurm_driver"): slurm_schema,
        Optional("htcondor_driver"): htcondor_schema,
        Optional("notify", default=[]): [{"name": str, Optional(object): object}],
        Optional(object): object,
    }
)


class Notifier:
    name: typing.Optional[str] = None
    args: typing.Optional[typing.Dict[str, typing.Any]] = None
    _notifier: notifiers.core.Provider

    def __init__(self, name: str, **kwargs: typing.Any):
        from .logger import logger

        self.logger = logger
        self.name = name
        self._notifier = notifiers.get_notifier(self.name)
        self._kwargs = kwargs

    def notify(
        self,
        message: str,
        title: typing.Optional[str] = None,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> notifiers.core.Response:
        self.logger.debug("Sending notification '%s' via %s", message, self.name)

        kwargs = kwargs.copy()
        kwargs["message"] = f"{socket.gethostname()}: {message}"
        if title is not None:
            if "title" in self._notifier.schema["properties"]:
                kwargs["title"] = title
            elif "subject" in self._notifier.schema["properties"]:
                kwargs["subject"] = title
            else:
                kwargs["message"] = title + ":\n" + kwargs["message"]

        return self._notifier.notify(**self._kwargs, **kwargs)


class NotificationManager:
    notifiers: typing.List[Notifier]

    def __init__(self, config: "Config"):
        from .logger import logger

        self.logger = logger

        self.notifiers = []
        if "notify" in config.data and config.data["notify"] is not None:
            for item in config.data["notify"]:
                name = item["name"]
                args = dict(item)
                del args["name"]
                self.notifiers.append(Notifier(name=name, **args))

    def notify(
        self,
        message: str,
        title: typing.Optional[str] = None,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.List[notifiers.core.Response]:  # type: ignore
        self.logger.debug("%d notifiers configured", len(self.notifiers))

        responses: typing.List[notifiers.core.Response] = []
        for notifier in self.notifiers:
            responses.append(notifier.notify(message, title, *args, **kwargs))
        return responses


class Config:
    def __init__(
        self, data: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> None:
        if data is not None:
            self.data = data
        else:
            with open(CONFIG_FILE) as f:
                self.data = yaml.safe_load(f)

        self.data = config_schema.validate(self.data)

        self.notifications = NotificationManager(self)

    def __getattr__(self, key: str) -> typing.Any:
        return self.data[key]



import datetime
import itertools
import json
import os
import re
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import (
    Any,
    Iterator,
    Iterable,
    Optional,
    Union,
    List,
    Dict,
    Sequence,
    cast,
    Collection,
)

import sh
from jinja2 import Environment, DictLoader

from .batch_driver_base import BatchDriverBase
from . import InvalidJobStatus
from ..logger import logger
from ..config import Config
from ..db import database
from ..model import Job, Folder
from ..util import make_executable, parse_timedelta
from .driver_base import DriverBase, checked_job


class HTCondorAccountingItem:
    job_id: int
    condor_status: int
    exit_code: int
    status: Job.Status

    def __init__(self, job_id: int, status: Job.Status, exit_code: int):
        self.job_id = job_id
        self.status = status
        self.exit_code = exit_code

    @classmethod
    def from_parts(
        cls, job_id: int, condor_status: int, exit_code: int
    ) -> "HTCondorAccountingItem":
        # see http://pages.cs.wisc.edu/~adesmet/status.html

        if condor_status == 0:
            # 0 Unexpanded U
            status = Job.Status.SUBMITTED
        elif condor_status == 1:
            # 1 Idle I
            status = Job.Status.SUBMITTED
        elif condor_status == 2:
            # 2 Running R
            status = Job.Status.RUNNING
        elif condor_status == 3:
            # 3 Removed X
            status = Job.Status.FAILED
        elif condor_status == 4:
            # 4 Completed C
            status = Job.Status.COMPLETED
        elif condor_status == 5:
            # 5 Held H
            status = Job.Status.FAILED
        elif condor_status == 6:
            # 6 Submission_err E
            status = Job.Status.FAILED
        else:
            status = Job.Status.UNKNOWN

        if status == Job.Status.COMPLETED:
            # at scheduler level, this is completed, might have exited with failure though
            if exit_code != 0 and exit_code != -1:
                status = Job.Status.FAILED
        return cls(job_id, status, exit_code)

    def __repr__(self) -> str:
        return f"HTCondorAI<{self.job_id}, {repr(self.status)}, {self.exit_code}>"

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, self.__class__)
        return (
            self.job_id == other.job_id
            and self.status == other.status
            and self.exit_code == other.exit_code
        )


class HTCondorInterface(ABC):
    @abstractmethod
    def condor_submit(self, job: Job) -> int:
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def condor_q(self) -> Iterator[HTCondorAccountingItem]:
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def condor_history(self) -> Iterator[HTCondorAccountingItem]:
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def condor_rm(self, job: "Job") -> None:
        raise NotImplementedError()  # pragma: no cover


class ShellHTCondorInterface(HTCondorInterface):
    _condor_submit: Optional[sh.Command] = None
    _condor_q: Optional[sh.Command] = None
    _condor_history: Optional[sh.Command] = None
    _condor_rm: Optional[sh.Command] = None

    config: Dict[str, Any]

    subreg = re.compile(r".* (\d+)\.$")

    def __init__(self, config: Dict[str, Any]) -> None:  # pragma: no cover
        self.config = config
        self._condor_submit = sh.Command("condor_submit")
        self._condor_q = sh.Command("condor_q")
        self._condor_history = sh.Command("condor_history")
        self._condor_rm = sh.Command("condor_rm")

    def _parse_output(self, output: str) -> Iterator[HTCondorAccountingItem]:
        if output.strip() == "":
            return []
        data = json.loads(output)
        for item in data:
            job_id = item["ClusterId"]
            assert item["ProcId"] == 0, "Clusters with more than one jobs not supported"
            condor_status = item["JobStatus"]
            exit_code = item["ExitCode"] if "ExitCode" in item else -1
            yield HTCondorAccountingItem.from_parts(job_id, condor_status, exit_code)

    def condor_q(self) -> Iterator[HTCondorAccountingItem]:

        logger.debug("Getting job infos")

        args = ["-attributes", ",".join(["ClusterId", "ProcId", "JobStatus"]), "-json"]

        assert self._condor_q is not None
        return self._parse_output(str(self._condor_q(*args)))

    def condor_history(self) -> Iterator[HTCondorAccountingItem]:

        logger.debug("Getting job infos")

        args = [
            self.config["user"],
            "-attributes",
            ",".join(["ClusterId", "ProcId", "JobStatus", "ExitCode"]),
            "-json",
            "-limit",
            "10000",
        ]

        assert self._condor_history is not None
        return self._parse_output(str(self._condor_history(*args)))

    def condor_submit(self, job: Job) -> int:
        assert self._condor_submit is not None
        raw = self._condor_submit(job.data["batchfile"])
        logger.debug("condor_submit: %s", raw)
        m = self.subreg.search(str(raw))
        assert m is not None
        return int(m.group(1))

    def condor_rm(self, job: Job) -> None:
        assert self._condor_rm is not None
        res = self._condor_rm(job.batch_job_id)
        logger.debug("condor_rm: %s", res)


jobscript_tpl_str = """
#!/usr/bin/env bash

export KONG_JOB_ID={{internal_job_id}}
export KONG_JOB_OUTPUT_DIR={{output_dir}}
export KONG_JOB_LOG_DIR={{log_dir}}
export KONG_JOB_NPROC={{cores}}
export KONG_JOB_SCRATCHDIR=$_CONDOR_SCRATCH_DIR
export HTCONDOR_CLUSTER_ID=$(grep "^ClusterId" $_CONDOR_JOB_AD | cut -d= - -f2 | awk '{$1=$1};1')

mkdir -p $KONG_JOB_SCRATCHDIR

stdout={{stdout}}

({{command}}) > $stdout 2>&1
""".strip()

batchfile_tpl_str = """
universe = vanilla
log = {{htcondor_out}}
executable = {{jobscript}}
request_cpus = {{cores}}
request_memory = {{memory}}
batch_name = {{name}}
+MaxRuntime = {{walltime}}

{{submitfile_extra}}

queue 1
""".strip()  # noqa: W291, W293

env = Environment(
    loader=DictLoader({"batchfile": batchfile_tpl_str, "jobscript": jobscript_tpl_str})
)

batchfile_tpl = env.get_template("batchfile")
jobscript_tpl = env.get_template("jobscript")


class HTCondorDriver(BatchDriverBase):
    htcondor: HTCondorInterface

    def __init__(self, config: Config, htcondor: Optional[HTCondorInterface] = None):
        DriverBase.__init__(self, config)
        self.htcondor_config = self.config.data["htcondor_driver"]
        self.htcondor = htcondor or ShellHTCondorInterface(self.htcondor_config)

    def create_job(
        self,
        folder: "Folder",
        command: str,
        cores: int = 1,
        memory: int = 2000,
        universe: Optional[str] = None,
        name: Optional[str] = None,
        walltime: Union[timedelta, str] = timedelta(minutes=30),
    ) -> "Job":  # type: ignore

        if universe is None:
            universe = self.htcondor_config["default_universe"]

        job: Job = Job.create(
            folder=folder,
            batch_job_id=None,  # don't have one until submission
            command=command,
            driver=self.__class__,
            cores=cores,
            memory=memory,
        )

        if name is None:
            name = f"kong_job_{job.job_id}"

        # in job dir, create output dir
        output_dir = self.make_output_path(job)
        os.makedirs(output_dir, exist_ok=True)

        log_dir = self.make_log_path(job)
        os.makedirs(log_dir, exist_ok=True)

        stdout = os.path.abspath(os.path.join(log_dir, "stdout.txt"))
        htcondor_out = os.path.abspath(os.path.join(log_dir, "htcondor_out.txt"))

        batchfile = os.path.join(log_dir, "batchfile.sh")
        jobscript = os.path.join(log_dir, "jobscript.sh")

        if isinstance(walltime, str):
            norm_walltime = int(parse_timedelta(walltime).total_seconds())
        elif isinstance(walltime, timedelta):
            norm_walltime = int(walltime.total_seconds())
        else:
            raise ValueError("Walltime must be timedelta or string")

        job.data = dict(
            stdout=stdout,
            htcondor_out=htcondor_out,
            jobscript=jobscript,
            batchfile=batchfile,
            output_dir=output_dir,
            log_dir=log_dir,
            name=name,
            exit_code=0,
            universe=universe,
            walltime=norm_walltime,
            submitfile_extra=self.htcondor_config["submitfile_extra"],
        )
        job.save()

        values = dict(
            batchfile=batchfile,
            jobscript=jobscript,
            command=command,
            stdout=stdout,
            htcondor_out=htcondor_out,
            internal_job_id=job.job_id,
            log_dir=log_dir,
            output_dir=output_dir,
            cores=cores,
            memory=memory,
            universe=universe,
            name=name,
            walltime=norm_walltime,
            submitfile_extra=self.htcondor_config["submitfile_extra"],
        )

        batchfile_content = batchfile_tpl.render(**values)

        with open(batchfile, "w") as fh:
            fh.write(batchfile_content)

        jobscript_content = jobscript_tpl.render(**values)
        with open(job.data["jobscript"], "w") as fh:
            fh.write(jobscript_content)

        make_executable(job.data["jobscript"])

        job._driver_instance = self
        return job

    def bulk_sync_status(self, jobs: Collection["Job"]) -> Sequence["Job"]:
        logger.debug("Bulk sync status with %d jobs", len(jobs))
        for job in jobs:
            self._check_driver(job)

        now = datetime.datetime.now()

        def proc() -> Iterable[Job]:
            job_not_found = 0
            for item in itertools.chain(
                self.htcondor.condor_q(), self.htcondor.condor_history()
            ):
                job = Job.get_or_none(batch_job_id=item.job_id)
                if job is None:
                    job_not_found += 1
                    continue
                job.status = item.status
                job.data["exit_code"] = item.exit_code
                job.updated_at = now
                yield job
            if job_not_found > 0:
                logger.warning(
                    "Tried to fetch %d slurm jobs which where not found in the database",
                    job_not_found,
                )

        with database.atomic():
            Job.bulk_update(
                proc(),
                fields=[Job.data, Job.status, Job.updated_at],
                batch_size=self.batch_size,
            )
        # reload updated jobs
        ids = [j.job_id for j in jobs]
        logger.debug(
            "Going to reload %d jobs, batch size %d", len(ids), self.select_batch_size
        )

        fetched: List[Job] = list(
            Job.bulk_select(Job.job_id, ids, batch_size=self.select_batch_size)
        )
        return cast(Sequence[Job], fetched)

    @checked_job
    def submit(self, job: "Job", save: bool = True) -> None:
        if job.status != Job.Status.CREATED:
            raise InvalidJobStatus(f"Cannot submit job {job} in status {job.status}")
        job.batch_job_id = str(self.htcondor.condor_submit(job))
        job.status = Job.Status.SUBMITTED

        if save:
            job.save()

    @checked_job
    def kill(self, job: "Job", save: bool = True) -> None:
        if job.status in (Job.Status.CREATED, Job.Status.UNKNOWN):
            logger.debug("Job %s in %s, simply setting to failed", job, job.status)
            job.status = Job.Status.FAILED
        elif job.status in (Job.Status.RUNNING, Job.Status.SUBMITTED):
            self.htcondor.condor_rm(job)
            job.status = Job.Status.FAILED
        else:
            logger.debug("Job %s in %s, do nothing")
        if save:
            job.save()



import os
import tempfile
import time

import pytest
from unittest import mock
from unittest.mock import Mock, ANY, MagicMock
import peewee as pw
from conftest import skip_lxplus

from kong.model import Folder, Job
from kong.repl import Repl, complete_path
import kong

import logging

kong.logger.logger.setLevel(logging.DEBUG)


@pytest.fixture
def repl(state):
    return Repl(state)


def test_ls(tree, state, repl, capsys, sample_jobs, monkeypatch):
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j.job_id}" in out for j in state.cwd.jobs)

    state.mkdir("f4")
    repl.onecmd("ls f4")

    j1 = Job.get(1)
    j1.batch_job_id = None
    j1.save()
    state.mv(j1, "f1")

    repl.onecmd("ls")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j}" in out for j in [2, 3])

    repl.onecmd("ls --recursive")
    out, err = capsys.readouterr()
    #assert all(f.name in out for f in state.cwd.children)
    all_jobs = [j.job_id for j in Job.select()]
    assert all(f"{j}" in out for j in all_jobs)

    repl.do_ls("/nope")
    out, err = capsys.readouterr()
    assert "not exist" in out

    state.cwd = Folder.find_by_path(state.cwd, "/f2")
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j.job_id}" in out for j in state.cwd.jobs)

    repl.onecmd("ls --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out

    with monkeypatch.context() as m:

        job = state.create_job(command="sleep 1")
        job.batch_job_id = None

        m.setattr(state, "refresh_jobs", Mock())
        m.setattr(state, "ls", Mock(return_value=([], [job])))

        repl.onecmd("ls -R /")


@skip_lxplus
def test_ls_refresh(repl, state, capsys, sample_jobs, monkeypatch):
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert "CREATED" in out

    for job in sample_jobs:
        job.submit()

    time.sleep(0.2)

    # without refresh
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert "CREATED" not in out
    assert "COMPLETED" not in out
    assert "SUBMITTED" in out

    # with refresh
    repl.do_ls(". --refresh")
    out, err = capsys.readouterr()
    assert "SUBMITTED" not in out
    assert "COMPLETED" in out

    with monkeypatch.context() as m:
        mock = Mock()
        m.setattr(state, "refresh_jobs", mock)
        repl.onecmd("ls -r /")
        assert mock.call_count == 1  # called once


def test_complete_funcs(state, tree, repl, monkeypatch):
    root = Folder.get_root()
    cmpl = Mock()
    monkeypatch.setattr("kong.repl.complete_path", cmpl)

    for c in [
        "ls",
        "mkdir",
        "mv",
        "rm",
        "cd",
        "submit_job",
        "kill_job",
        "info",
        "resubmit_job",
    ]:
        func = getattr(repl, f"complete_{c}")

        func("hurz", "ls hurz", 4, 6)
        cmpl.assert_called_once_with(root, "hurz")
        cmpl.reset_mock()

        func("hurz/", "ls hurz/", 8, 8)
        cmpl.assert_called_once_with(root, "hurz/")
        cmpl.reset_mock()

        func("", "ls hurz/ blubb", 8, 8)
        cmpl.assert_called_once_with(root, "hurz/")
        cmpl.reset_mock()

        func("blubb", "ls hurz/ blubb", 9, 13)
        cmpl.assert_called_once_with(root, "blubb")
        cmpl.reset_mock()

        func("blubb", "ls hurz/ blubb/", 15, 15)
        cmpl.assert_called_once_with(root, "blubb/")
        cmpl.reset_mock()

        func("blu", "ls hurz/blu", 8, 10)
        cmpl.assert_called_once_with(root, "hurz/blu")
        cmpl.reset_mock()


def test_complete_path(state, tree, repl):
    root = Folder.get_root()
    alts = complete_path(root, "f")
    assert alts == ["f1/", "f2/", "f3/"]

    alts = complete_path(root, "f1")
    assert alts == ["f1/"]

    alts = complete_path(root, "f2/")
    assert alts == ["alpha/", "beta/", "gamma/"]

    state.cwd = Folder.find_by_path(state.cwd, "/f2")

    alts = complete_path(root.subfolder("f2"), "a")
    assert alts == ["alpha/"]


def test_mkdir(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()
    sub = root.add_folder("sub")

    for cwd in [root, sub]:
        state.cwd = cwd

        assert cwd.subfolder("alpha") is None
        repl.do_mkdir("alpha")
        out, err = capsys.readouterr()
        alpha = cwd.subfolder("alpha")
        assert alpha is not None

        # one down
        assert alpha.subfolder("beta") is None
        repl.do_mkdir("alpha/beta")
        out, err = capsys.readouterr()
        beta = alpha.subfolder("beta")
        assert beta is not None

        # cannot create outside of root
        repl.do_mkdir("../nope")
        if cwd == root:
            out, err = capsys.readouterr()
            assert "annot create" in out and "../nope" in out
        else:
            assert root.subfolder("nope") is not None

        # cannot create again
        repl.do_mkdir("alpha")
        out, err = capsys.readouterr()
        assert "alpha" in out

        # cannot create in nonexistant
        repl.do_mkdir("omega/game")
        out, err = capsys.readouterr()
        assert "omega/game" in out and "annot create" in out

        state.cwd = beta
        assert cwd.subfolder("gamma") is None
        repl.do_mkdir("../../gamma")
        out, err = capsys.readouterr()
        gamma = cwd.subfolder("gamma")
        assert gamma is not None

        # force integrity error
        with monkeypatch.context() as m:
            m.setattr(state, "mkdir", Mock(side_effect=pw.IntegrityError))
            repl.onecmd("mkdir hurz")
            out, err = capsys.readouterr()
            assert "already exists" in out

    assert Repl.do_mkdir.__doc__ is not None


def test_mkdir_create_parents(state, repl, capsys):
    root = Folder.get_root()

    repl.onecmd("mkdir /a1/b2/c3/d4")
    out, err = capsys.readouterr()
    assert "Cannot create folder" in out
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3/d4") is None

    repl.onecmd("mkdir -p /a1/b2/c3/d4")
    out, err = capsys.readouterr()
    assert Folder.find_by_path(state.cwd, "/a1") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3/d4") is not None


def test_cd(state, repl, db, capsys):
    root = Folder.get_root()
    assert state.cwd == root

    repl.do_cd("nope")
    out, err = capsys.readouterr()
    assert "not exist" in out, "nope" in out
    assert state.cwd == root

    nope = root.add_folder("nope")
    repl.do_cd("nope")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("")
    out, err = capsys.readouterr()
    assert state.cwd == root

    repl.do_cd("..")
    out, err = capsys.readouterr()
    assert "not exist" in out
    assert state.cwd == root

    repl.do_cd("../nope")
    out, err = capsys.readouterr()
    assert "not exist" in out
    assert state.cwd == root

    more = root.add_folder("more")
    another = nope.add_folder("another")

    repl.do_cd("/nope")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("/nope/another")
    out, err = capsys.readouterr()
    assert state.cwd == another

    repl.do_cd("/../")
    out, err = capsys.readouterr()
    assert state.cwd == another

    repl.do_cd("..")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("/more")
    out, err = capsys.readouterr()
    assert state.cwd == more


def test_mv_folder(state, repl, capsys):
    root = Folder.get_root()

    repl.onecmd("mv --help")
    out, err = capsys.readouterr()


    f1, f2, f3, f4, f5 = [root.add_folder(n) for n in ("f1", "f2", "f3", "f4", "f5")]

    assert len(root.children) == 5
    assert len(f2.children) == 0

    # actual move
    repl.onecmd("mv f1 f2")
    assert len(root.children) == 4
    f2.reload()
    assert len(f2.children) == 1 and f2.children[0] == f1
    f1.reload()
    assert f1.parent == f2
    out, err = capsys.readouterr()

    # rename f3 -> f3x
    repl.onecmd("mv f3 f3x")
    out, err = capsys.readouterr()
    f3.reload()
    assert len(root.children) == 4
    assert f3.name == "f3x"

    # another move
    repl.onecmd("mv f3x f4")
    assert len(f4.children) == 1 and f4.children[0] == f3
    f3.reload()
    assert f3.parent == f4
    assert f3.name == "f3x"
    out, err = capsys.readouterr()

    # move rename at the same time
    repl.onecmd("cd f2")
    repl.onecmd("mv ../f5 ../f4/f5x")
    out, err = capsys.readouterr()
    f5.reload()
    assert len(f4.children) == 2
    assert f5.name == "f5x"
    assert f5.parent == f4

    # try move to nonexistant
    repl.onecmd("cd /")
    repl.onecmd("mv f2/f1 /nope/blub")
    out, err = capsys.readouterr()
    assert "/nope" in out and "not exist" in out

    # try to move nonexistant
    repl.onecmd("mv ../nope f1")
    out, err = capsys.readouterr()
    assert "../nope" in out and "No such" in out


def test_mv_job(state, repl, capsys):
    root = Folder.get_root()

    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]

    assert len(root.children) == 2

    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(root.jobs) == 5

    repl.onecmd(f"mv {j1.job_id} f1")
    j1.reload()
    assert j1.folder == f1
    assert len(f1.jobs) == 1
    assert len(root.jobs) == 4
    out, err = capsys.readouterr()

    repl.onecmd(f"mv {j2.job_id} f2")
    j2.reload()
    assert j2.folder == f2
    assert len(f2.jobs) == 1
    assert len(root.jobs) == 3
    out, err = capsys.readouterr()

    state.cd(f2)
    repl.onecmd(f"mv {j3.job_id} .")
    j3.reload()
    assert j3.folder == f2
    out, err = capsys.readouterr()

    repl.onecmd(f"mv {j2.job_id} ..")
    j2.reload()
    assert j2.folder == root
    out, err = capsys.readouterr()

    repl.onecmd(f"mv ../{j4.job_id} ../f1")
    j4.reload()
    assert j4.folder == f1
    out, err = capsys.readouterr()

    state.cd(root)

    # renaming does not work
    repl.onecmd(f"mv {j5.job_id} 42")
    out, err = capsys.readouterr()
    assert "42" in out and "not exist" in out


def test_mv_bulk_job(state, repl):
    root = Folder.get_root()

    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]
    assert len(root.children) == 2

    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(root.jobs) == 5

    repl.onecmd("mv * f1")
    assert len(root.jobs) == 0 and len(f1.jobs) == 5
    assert len(root.children) == 1  # Didn't move the folders
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f1

    state.cwd = f2
    repl.onecmd("mv ../f1/* .")
    assert len(f1.jobs) == 0 and len(f2.jobs) == 5
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f2


def test_mv_bulk_folder(state, repl):
    root = Folder.get_root()

    r1, r2 = [root.add_folder(n) for n in ("r1", "r2")]

    folders = [r1.add_folder(f"f{n}") for n in range(5)]
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0

    repl.onecmd("mv r1/* r2")
    assert len(r1.children) == 0
    assert len(r2.children) == len(folders)
    for f in folders:
        f.reload()
        assert f.parent == r2

    state.cwd = r1
    repl.onecmd("mv ../r2/* .")
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0
    for f in folders:
        f.reload()
        assert f.parent == r1


def test_mv_error(state, repl, capsys, monkeypatch):
    repl.onecmd("mv --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_rm(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()

    repl.do_rm("../nope")
    out, err = capsys.readouterr()
    assert "not exist" in out

    repl.do_rm("/")
    out, err = capsys.readouterr()
    assert "annot delete" in out, "root" in out

    root.add_folder("alpha")
    with monkeypatch.context() as m:
        confirm = Mock(return_value=False)
        m.setattr("click.confirm", confirm)
        repl.onecmd("rm -R alpha")
        confirm.assert_called_once()

    assert root.subfolder("alpha") is not None
    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.onecmd("rm -R alpha")
        confirm.assert_called_once()
    assert root.subfolder("alpha") is None
    out, err = capsys.readouterr()
    assert len(out) > 0


def test_rm_job(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()
    j1 = state.default_driver.create_job(command="sleep 1", folder=root)
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    assert Job.get_or_none(job_id=j1.job_id) is not None

    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.do_rm(str(j1.job_id))
        confirm.assert_called_once()

    out, err = capsys.readouterr()
    assert len(root.jobs) == 0
    assert Job.get_or_none(job_id=j1.job_id) is None

    # works in other cwd too
    alpha = root.add_folder("alpha")
    j2 = state.default_driver.create_job(command="sleep 1", folder=alpha)
    assert j1.job_id != j2.job_id
    assert Job.get_or_none(job_id=j2.job_id) is not None
    assert len(alpha.jobs) == 1 and alpha.jobs[0] == j2
    assert state.cwd == root
    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.do_rm(str(j2.job_id))
        confirm.assert_called_once()
    out, err = capsys.readouterr()
    assert Job.get_or_none(job_id=j2.job_id) is None
    assert len(alpha.jobs) == 0


def test_cwd(state, repl, tree, capsys):
    root = tree
    repl.onecmd("cwd")
    out, err = capsys.readouterr()
    assert out.strip() == "/"

    state.cwd = root / "f1"
    repl.onecmd("cwd")
    out, err = capsys.readouterr()
    assert out.strip() == "/f1"

    state.cwd = root / "f2" / "gamma"
    repl.onecmd("cwd")
    out, err = capsys.readouterr()
    assert out.strip() == "/f2/gamma"


def test_wait(repl, monkeypatch):
    state = Mock()
    state.wait = Mock(return_value=iter([]))
    monkeypatch.setattr(repl, "state", state)

    repl.onecmd("wait * --no-notify --recursive --poll-interval 50")

    state.wait.assert_called_once_with(
        "*", notify=False, recursive=True, poll_interval=50
    )


def test_exit(repl):
    assert repl.do_exit("") == True
    assert repl.do_EOF("") == True


def test_preloop(repl, monkeypatch):
    m = Mock()
    monkeypatch.setattr("readline.read_history_file", m)
    monkeypatch.setattr("os.path.exists", Mock(return_value=True))
    repl.preloop()
    m.assert_called_once()
    monkeypatch.setattr("os.path.exists", Mock(return_value=False))
    repl.preloop()


def test_postloop(state, repl, monkeypatch):
    repl.postloop()


def test_precmd(repl):
    assert repl.precmd("whatever") == "whatever"
    assert repl.precmd("") == ""


def test_onecmd(repl, monkeypatch, capsys):
    set_length = Mock()
    write = Mock()
    monkeypatch.setattr("readline.set_history_length", set_length)
    monkeypatch.setattr("readline.write_history_file", write)
    m = Mock(return_value="ok")
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    assert repl.onecmd("whatever") == "ok"
    set_length.assert_called_once_with(repl.state.config.history_length)
    write.assert_called_once()
    m.assert_called_once()

    m = Mock(side_effect=TypeError("MESSAGE"))
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    repl.onecmd("whatever")
    out, err = capsys.readouterr()
    assert "MESSAGE" in out
    m = Mock(side_effect=RuntimeError())
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    repl.onecmd("whatever")  # swallows other exceptions


def test_cmdloop(repl, monkeypatch, capsys):
    m = Mock(return_value="ok")
    monkeypatch.setattr("cmd.Cmd.cmdloop", m)
    repl.cmdloop()
    m.assert_called_once()

    m = Mock(side_effect=[KeyboardInterrupt(), "ok"])
    monkeypatch.setattr("cmd.Cmd.cmdloop", m)
    repl.cmdloop()
    out, err = capsys.readouterr()
    m.assert_called()
    assert m.call_count == 2
    assert "^C" in out


def test_emptyline(repl):
    repl.emptyline()


def test_create_job(repl, state, tree, capsys):
    root = tree

    repl.do_create_job("")
    out, err = capsys.readouterr()
    assert "provide a command" in out

    cmd = "sleep 1"
    repl.do_create_job(f"{cmd}")

    assert len(root.jobs) == 1
    j1 = root.jobs[-1]
    assert j1.command == cmd
    assert j1.status == Job.Status.CREATED
    out, err = capsys.readouterr()
    assert "reated" in out
    assert str(j1.job_id) in out
    assert j1.batch_job_id in out

    cores = 16
    repl.do_create_job(f"-c {cores} '{cmd}'")
    assert len(root.jobs) == 2
    j2 = root.jobs[-1]
    assert j2.command == cmd
    assert j2.status == Job.Status.CREATED
    assert j2.cores == cores
    out, err = capsys.readouterr()
    assert str(j2.job_id) in out
    assert j2.batch_job_id in out

    repl.do_create_job(f"--cores {cores} -- {cmd}")
    assert len(root.jobs) == 3
    j3 = root.jobs[-1]
    assert j3.command == cmd
    assert j3.status == Job.Status.CREATED
    assert j3.cores == cores
    out, err = capsys.readouterr()
    assert str(j3.job_id) in out
    assert j3.batch_job_id in out

    repl.do_create_job("--help")
    out, err = capsys.readouterr()
    assert "Usage" in out

    repl.onecmd("create_job --nope 5 sleep 2")  # wrong option --core
    out, err = capsys.readouterr()
    assert "no such" in out

    repl.onecmd("create_job -- exe --and --some arguments --and options")
    out, err = capsys.readouterr()

    j4 = root.jobs[-1]
    assert j4.command == "exe --and --some arguments --and options"


@skip_lxplus
def test_submit_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    value = "VALUE VALUE VALUE"
    cmd = f"sleep 0.3; echo '{value}'"

    monkeypatch.setattr("click.confirm", Mock(return_value=True))

    repl.do_create_job(cmd)
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)
    assert j1.status == Job.Status.CREATED

    repl.do_submit_job(f"{j1.job_id}")
    j1.reload()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING
    time.sleep(0.3)
    assert j1.get_status() == Job.Status.COMPLETED

    out, err = capsys.readouterr()

    repl.onecmd("submit_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_kill_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    repl.do_create_job("sleep 1")
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)

    assert j1.status == Job.Status.CREATED
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING

    monkeypatch.setattr("click.confirm", Mock(return_value=True))
    repl.do_kill_job(str(j1.job_id))
    out, err = capsys.readouterr()
    assert j1.get_status() == Job.Status.FAILED

    repl.onecmd("kill_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_resubmit_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    repl.do_create_job("sleep 1")
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)

    assert j1.status == Job.Status.CREATED
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING

    monkeypatch.setattr("click.confirm", Mock(return_value=True))
    repl.do_kill_job(str(j1.job_id))

    assert j1.get_status() == Job.Status.FAILED

    repl.do_resubmit_job(str(j1.job_id))
    out, err = capsys.readouterr()
    j1.reload()
    assert j1.status == Job.Status.SUBMITTED

    repl.onecmd("resubmit_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


@skip_lxplus
def test_status_update(repl, state, capsys):
    root = Folder.get_root()

    repl.onecmd("status")
    out, err = capsys.readouterr()
    assert "usage" in out

    repl.onecmd("status 42")
    out, err = capsys.readouterr()
    assert "not find" in out

    j1 = state.create_job(command="sleep 0.2")

    def update():
        repl.onecmd(f"update {j1.job_id}")
        out, err = capsys.readouterr()

    update()
    repl.onecmd(f"status {j1.job_id}")
    out, err = capsys.readouterr()
    assert "CREATED" in out

    j1.submit()

    time.sleep(0.1)

    repl.onecmd(f"status {j1.job_id}")
    out, err = capsys.readouterr()
    assert "SUBMITTED" in out

    update()
    repl.onecmd(f"status {j1.job_id}")
    out, err = capsys.readouterr()
    assert "RUNNING" in out

    time.sleep(0.2)

    repl.onecmd(f"status {j1.job_id}")
    out, err = capsys.readouterr()
    assert "RUNNING" in out

    repl.onecmd(f"status -r {j1.job_id}")
    out, err = capsys.readouterr()
    assert "COMPLETED" in out

    repl.onecmd("update --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_info(state, repl, capsys, monkeypatch):
    job = state.create_job(command="sleep 1")

    repl.onecmd("info 1")
    out, err = capsys.readouterr()
    assert str(job.job_id) in out
    assert job.batch_job_id in out
    for k, v in job.data.items():
        assert k in out
        assert v in out

    with monkeypatch.context() as m:
        refresh = Mock()
        m.setattr(state, "refresh_jobs", refresh)
        repl.onecmd("info 1 -r")
        out, err = capsys.readouterr()
        refresh.assert_called_once()

    command_string = "X" * 600

    long_job = state.create_job(command=command_string)

    repl.onecmd("info 2")
    out, err = capsys.readouterr()
    assert not command_string in out

    repl.onecmd("info 2 --full")
    out, err = capsys.readouterr()
    assert command_string in out


def test_tail(state, repl, capsys, monkeypatch):
    job = state.create_job(command="sleep 1")

    with monkeypatch.context() as m:
        tail = Mock(return_value=["LINENO1", "LINENO2"])
        m.setattr("sh.tail", tail)
        m.setattr("time.sleep", Mock())
        m.setattr("os.path.exists", Mock(side_effect=[False, False, True]))
        spinner = MagicMock()
        m.setattr("kong.repl.Spinner", spinner)
        repl.onecmd(f"tail {job.job_id}")
        out, err = capsys.readouterr()
        spinner.assert_called_once()

    with monkeypatch.context() as m:
        tail = Mock(return_value=["LINENO1", "LINENO2"])
        m.setattr("sh.tail", tail)
        m.setattr("os.path.exists", Mock(side_effect=[True]))
        spinner = MagicMock()
        m.setattr("kong.repl.Spinner", spinner)
        repl.onecmd(f"tail {job.job_id}")
        out, err = capsys.readouterr()
        assert spinner.call_count == 0

    with monkeypatch.context() as m:
        repl.onecmd(f"tail --nope")
        out, err = capsys.readouterr()
        assert "no such option" in out


def test_less(state, repl, capsys, monkeypatch):
    job = state.create_job(command="sleep 1")

    content = "SOMECONTENT: BLABLBALBALBALBLA\nNEWLINE"
    with tempfile.NamedTemporaryFile("wt") as f:
        f.write(content)
        f.flush()
        job.data["stdout"] = f.name
        job.save()

        lines = []

        def agg(it):
            nonlocal lines
            lines = list(it)

        with monkeypatch.context() as m:
            pager = Mock(side_effect=agg)
            m.setattr("click.echo_via_pager", pager)
            repl.onecmd(f"less {job.job_id}")
            out, err = capsys.readouterr()
            pager.assert_called_once()
            assert "".join(lines) == content

    repl.onecmd(f"less --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out



import os

import pytest
import peewee as pw

from kong.db import database
from kong.model import Folder

import logging

logging.getLogger("kong").setLevel(logging.DEBUG)


def test_create(db):
    root = Folder.get_root()
    f1 = Folder.create(name="hallo", parent=root)

    assert f1.parent is root
    assert len(root.children) == 1

    f2 = Folder.create(name="number2", parent=f1)
    assert len(root.children) == 1
    assert f2.name == "number2"
    assert f2.parent == f1
    assert f2.parent_id == f1.folder_id

    with pytest.raises(AssertionError):
        Folder.create(name=".")
    with pytest.raises(AssertionError):
        Folder.create(name="..")
    with pytest.raises(AssertionError):
        Folder.create(name="/")

    f3 = root.add_folder("f3")
    assert f3.name == "f3"
    assert f3.parent == root
    assert len(root.children) == 2


def test_cannot_be_own_parent(db):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    assert f1.parent == root

    with pytest.raises(pw.IntegrityError):
        f1.parent = f1
        f1.save()
    f1.reload()
    assert f1.parent == root


def test_create_name_unique(db):
    root = Folder.get_root()
    f1 = root.add_folder("f1")

    with pytest.raises(pw.IntegrityError):
        Folder.create(name="f1", parent=root)
    with pytest.raises(pw.IntegrityError):
        root.add_folder("f1")

    # another one is fine
    assert root.add_folder("f2").name == "f2"
    # one further down is also fine
    assert f1.add_folder("f1").name == "f1"


def test_create_name_validate(db):
    root = Folder.get_root()
    with pytest.raises(AssertionError):
        Folder.create(name="123", parent=root)


def test_get_subfolder(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = root.add_folder("f2")
    f2_f1 = f2.add_folder("f1")

    assert root.subfolder("f1") == f1
    assert root.subfolder("f2") == f2
    assert f2.subfolder("f1") == f2_f1


def test_get_root_creates(db):
    assert Folder.get_or_none(name="root", parent=None) is None

    root = Folder.get_root()
    assert root.name == "root"
    assert root.parent is None

    assert Folder.get_or_none(name="root", parent=None) is not None


def test_only_one_root(db):
    root = Folder.get_root()
    # try to create another one

    with pytest.raises(AssertionError):
        Folder.create(name="something else")
    with pytest.raises(AssertionError):
        Folder.create(name="something else", parent=None)

    # make sure it wasnt created
    assert Folder.select().where(Folder.parent == None).count() == 1


def test_path(db):
    root = Folder.get_root()
    assert root.path == "/"

    f1 = Folder.create(name="f1", parent=root)
    assert f1.path == "/f1"
    f2 = Folder.create(name="f2", parent=f1)
    assert f2.path == "/f1/f2"
    f3 = Folder.create(name="f3", parent=f2)
    assert f3.path == "/f1/f2/f3"
    f4 = Folder.create(name="f4", parent=f3)
    assert f4.path == "/f1/f2/f3/f4"

    # move stuff around
    f4.parent = f2
    f4.save()
    assert f4.path == "/f1/f2/f4"


def test_find_by_path(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")
    f3 = f2.add_folder("f3")
    f4 = f2.add_folder("f4")

    for f in [root, f1, f2, f3, f4]:
        # absolute paths work regardless of cwd
        assert Folder.find_by_path(f, "/") == root
        assert Folder.find_by_path(f, "/f1") == f1
        assert Folder.find_by_path(f, "/f1/f2") == f2
        assert Folder.find_by_path(f, "/f1/f2/f3") == f3
        assert Folder.find_by_path(f, "/f1/f2/f4") == f4
        # self referential paths work everywhere
        assert Folder.find_by_path(f, ".") == f
        assert Folder.find_by_path(f, f.path) == f
        assert Folder.find_by_path(f, f.path + "/") == f

        # for root, it's None
        assert Folder.find_by_path(f, "..") == f.parent

        assert Folder.find_by_path(f, "nope") is None
        assert Folder.find_by_path(f, "../nope") is None

    assert Folder.find_by_path(root, "f1") == f1
    assert Folder.find_by_path(f1, "../") == root
    assert Folder.find_by_path(f1, "../f1") == f1
    assert Folder.find_by_path(f1, "../f1/f2") == f2
    assert Folder.find_by_path(f1, "f2") == f2
    assert Folder.find_by_path(f2, "f3") == f3
    assert Folder.find_by_path(f2, "f4") == f4
    assert Folder.find_by_path(f3, "../f4") == f4
    assert Folder.find_by_path(f4, "../f3") == f3


def test_fancy_operator(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")
    f3 = f2.add_folder("f3")
    f4 = f2.add_folder("f4")

    assert root / "f1" == f1
    assert root / "f1" / "f2" == f2
    assert root / "f1" / "f2" / "f3" == f3
    assert root / "f1" / "f2" / "f4" == f4


def test_jobs_recursive(db, state):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")

    with state.pushd(f1):
        j1 = state.create_job(command="sleep 1")
    with state.pushd(f2):
        j2 = state.create_job(command="sleep 1")

    jobs = root.jobs_recursive()
    assert len(jobs) == 2
    assert all(a == b for a, b in zip(jobs, [j1, j2]))



import os
import stat
from datetime import timedelta
from io import StringIO
from unittest.mock import Mock, ANY

import click
import pytest

from kong.util import (
    strip_colors,
    ljust,
    rjust,
    format_timedelta,
    parse_timedelta,
    shorten,
    shorten_path,
    make_executable,
    is_executable,
    rmtree,
    chunks,
    Spinner,
    Progress,
)


def test_strip_colors():
    s = "abcdef"
    assert strip_colors(s) == s

    assert strip_colors(click.style(s, fg="red")) == s


def test_ljust():
    s = "abcdlkjhdlfkjh"
    assert ljust(s, 30) == s.ljust(30)
    s_style = click.style(s, fg="red")
    assert ljust(s_style, 30)[-(30 - len(s)) :] == " " * (30 - len(s))


def test_rjust():
    s = "abcdlkjhdlfkjh"
    assert rjust(s, 30) == s.rjust(30)
    s_style = click.style(s, fg="red")
    assert rjust(s_style, 30)[: (30 - len(s))] == " " * (30 - len(s))


def test_format_timedelta():
    fmt = format_timedelta
    assert fmt(timedelta(seconds=30)) == "00:00:30"
    assert fmt(timedelta(minutes=42)) == "00:42:00"
    assert fmt(timedelta(minutes=42, seconds=14)) == "00:42:14"
    assert fmt(timedelta(hours=8)) == "08:00:00"
    assert fmt(timedelta(hours=99)) == "99:00:00"
    assert fmt(timedelta(days=3)) == f"{3*24}:00:00"
    with pytest.raises(ValueError):
        fmt(timedelta(days=5))
    assert fmt(timedelta(hours=99, minutes=59, seconds=59)) == "99:59:59"
    with pytest.raises(ValueError):
        fmt(timedelta(hours=100))
    assert fmt(timedelta(hours=8, minutes=6)) == "08:06:00"
    assert fmt(timedelta(hours=8, minutes=6, seconds=23)) == "08:06:23"


def test_parse_timedelta():
    ptd = parse_timedelta
    assert ptd("00:00:30") == timedelta(seconds=30)
    assert ptd("00:42:00") == timedelta(minutes=42)
    assert ptd("00:42:14") == timedelta(minutes=42, seconds=14)
    assert ptd("08:00:00") == timedelta(hours=8)
    assert ptd("99:00:00") == timedelta(hours=99)
    assert ptd(f"{3 * 24}:00:00") == timedelta(days=3)
    with pytest.raises(ValueError):
        ptd("100:00:00")
    assert ptd("99:59:59") == timedelta(hours=99, minutes=59, seconds=59)
    assert ptd("08:06:00") == timedelta(hours=8, minutes=6)
    assert ptd("08:06:23") == timedelta(hours=8, minutes=6, seconds=23)


def test_shorten():
    assert shorten("abcabcabcabc", 5) == "a...c"
    assert shorten("abcabcabcabc", 6) == "a...bc"
    assert shorten("abcabcabcabc", 7) == "ab...bc"
    assert shorten("abcdefghijklmnopqrstuvwxyz", 8) == "ab...xyz"
    assert shorten("abcdefghijklmnopqrstuvwxyz", 9) == "abc...xyz"
    assert shorten("abcabcabcabc", 12) == "abcabcabcabc"
    assert shorten("abcabcabcabc", 15) == "abcabcabcabc"
    with pytest.raises(ValueError):
        shorten("abcdefgh", 4)  # doesn't make sense


def test_shorten_path():
    assert (
        shorten_path("/a/very/long/path/with/many/segments") == "/a/v/l/p/w/m/segments"
    )
    assert shorten_path("a/very/long/path/with/many/segments") == "a/v/l/p/w/m/segments"

    assert (
        shorten_path("a/very/long/path/with/many/segments_is_very_long_too", 10)
        == "a/v/l/p/w/m/seg..._too"
    )


def test_make_executable(tmp_path):
    p = tmp_path.joinpath("test.sh")
    p.write_text("hallo")

    mode = os.stat(p).st_mode
    assert (mode & stat.S_IEXEC) == 0

    make_executable(p)
    mode = os.stat(p).st_mode
    assert (mode & stat.S_IEXEC) != 0


def test_is_executable(tmp_path):
    p = tmp_path.joinpath("test.sh")
    p.write_text("hallo")
    assert is_executable(p) == False

    os.chmod(p, os.stat(p).st_mode | stat.S_IEXEC)
    assert is_executable(p) == True


def test_rmtree(monkeypatch):
    with monkeypatch.context() as m:
        rmtree_mock = Mock(side_effect=OSError)
        m.setattr("shutil.rmtree", rmtree_mock)
        system = Mock()
        m.setattr("os.system", system)
        rmtree("whatever/blaaa")
        rmtree_mock.assert_called_once_with("whatever/blaaa")
        system.assert_called_once_with("rm -rf whatever/blaaa")

    with monkeypatch.context() as m:
        rmtree_mock = Mock()
        m.setattr("shutil.rmtree", rmtree_mock)
        system = Mock()
        m.setattr("os.system", system)
        rmtree("whatever/blaaa")
        rmtree_mock.assert_called_once_with("whatever/blaaa")
        assert system.call_count == 0


def test_chunks():
    l = [1, 2, 3, 4, 5, 6, 7]
    ch = list(chunks(l, 2))

    assert ch == [[1, 2], [3, 4], [5, 6], [7]]


def test_spinner(monkeypatch):

    with monkeypatch.context() as m:
        write = Mock()
        isatty = Mock(return_value=False)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("sys.stdout.write", write)
        with Spinner(text="blub"):
            pass
        write.assert_called_once_with("blub\n")

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with Spinner(text="blub"):
            pass

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 1
        assert HaloInstance.stop.call_count == 0
        assert HaloInstance.fail.call_count == 0
        Halo.assert_called_once()

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with pytest.raises(RuntimeError):
            with Spinner(text="blub", spinner="dots"):
                raise RuntimeError()

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 0
        assert HaloInstance.fail.call_count == 1
        Halo.assert_called_once_with("blub", spinner="dots")

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with pytest.raises(RuntimeError):
            with Spinner(text="blub", persist=False):
                raise RuntimeError()

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 1
        assert HaloInstance.fail.call_count == 0

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with Spinner(text="blub", persist=False):
            pass

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 1
        assert HaloInstance.fail.call_count == 0


def test_progress(monkeypatch):
    tqdm = Mock(return_value=[])
    monkeypatch.setattr("kong.util.tqdm", tqdm)
    for i in Progress(range(10)):
        pass
    assert tqdm.call_count == 1



import os
import logging

from unittest.mock import Mock

from kong.cli import main
import kong
from kong.model import Folder
from kong import logger


def test_verbosity(app_env, cli):
    app_dir, config_path, tmp_path = app_env

    result = cli.invoke(main, ["--version"])
    assert result.exception is None
    assert result.exit_code == 0
    assert logger.logger.getEffectiveLevel() == logging.WARNING
    assert logging.getLogger().getEffectiveLevel() == logging.WARNING

    result = cli.invoke(main, ["--version", "-v"])
    assert result.exit_code == 0
    assert result.exception is None
    assert logger.logger.getEffectiveLevel() == logging.INFO
    assert logging.getLogger().getEffectiveLevel() == logging.INFO

    result = cli.invoke(main, ["--version", "-vv"])
    assert result.exit_code == 0
    assert result.exception is None
    assert logger.logger.getEffectiveLevel() == logging.DEBUG
    assert logging.getLogger().getEffectiveLevel() == logging.INFO

    result = cli.invoke(main, ["--version", "-vvv"])
    assert result.exit_code == 0
    assert result.exception is None
    assert logger.logger.getEffectiveLevel() == logging.DEBUG
    assert logging.getLogger().getEffectiveLevel() == logging.DEBUG


def test_setup_implicit(app_env, db, cli, monkeypatch):
    app_dir, config_path, tmp_path = app_env
    print("APPDIR:", kong.config.APP_DIR)
    print("CONFDIR:", kong.config.CONFIG_FILE)

    repl = Mock()
    monkeypatch.setattr("kong.cli.Repl", repl)

    assert not os.path.exists(config_path)
    result = cli.invoke(main, ["-vv"], input="\n\n")
    assert result.exit_code == 0
    assert result.exception is None
    assert os.path.exists(config_path)
    cfg = kong.config.Config()
    assert cfg.default_driver == "kong.drivers.local_driver.LocalDriver"
    assert cfg.jobdir == os.path.join(app_dir, "jobdir")

    # assert db and root folder was created
    assert Folder.get_or_none(name="root", parent=None) is not None

    # run again
    result = cli.invoke(main, ["-vv"])


def test_setup_invalid_driver(app_env, db, cli):
    app_dir, config_path, tmp_path = app_env

    assert not os.path.exists(config_path)
    result = cli.invoke(main, [], input="NotADriver\n\n")
    assert result.exit_code == 1
    assert result.exception is not None


def test_setup_explicit(app_env, db, cli):
    app_dir, config_path, tmp_path = app_env

    assert not os.path.exists(config_path)
    result = cli.invoke(main, ["-vv", "setup"], input="\n\n")
    assert result.exit_code == 0
    assert result.exception is None
    assert os.path.exists(config_path)
    cfg = kong.config.Config()
    assert cfg.default_driver == "kong.drivers.local_driver.LocalDriver"
    assert cfg.jobdir == os.path.join(app_dir, "jobdir")

    # assert db and root folder was created
    assert Folder.get_or_none(name="root", parent=None) is not None


def test_repl_raises(app_env, db, cli, monkeypatch):
    app_dir, config_path, tmp_path = app_env
    cmdloop = Mock(side_effect=RuntimeError())
    monkeypatch.setattr("kong.repl.Repl.cmdloop", cmdloop)
    result = cli.invoke(main, [])
    assert result.exit_code == 1
    assert result.exception is not None


def test_interactive(app_env, cli, monkeypatch):
    mock = Mock()
    monkeypatch.setattr("IPython.embed", mock)

    cli.invoke(main, ["interactive"])
    mock.assert_called_once()




from typing import Any, Iterator


class RunningCommand:
    def __init__(self, *args: Any, **kwargs: Any): ...

    def __iter__(self) -> Iterator[str]: ...

class Command:
    def __init__(self, *args: Any, **kwargs: Any): ...

    def bake(self, *args: Any, **kwargs: Any) -> "Command": ...

    def __call__(self, *args: Any, **kwargs: Any) -> RunningCommand: ...

class SignalException_SIGTERM(BaseException): ...



from typing import TYPE_CHECKING, Any, List, ContextManager

if not TYPE_CHECKING:
    from playhouse.sqlite_ext import SqliteExtDatabase, AutoIncrementField
else:  # pragma: no cover

    class SqliteExtDatabase:
        def __init__(self, *args: Any) -> None:
            ...

        def init(self, *args: Any) -> None:
            ...

        def connect(self) -> None:
            ...

        def create_tables(self, tables: List[Any]) -> None:
            ...

        def atomic(self) -> ContextManager[None]:
            ...

    class AutoIncrementField:
        ...


database = SqliteExtDatabase(None)



import os
import datetime
from typing import Any, cast, Optional, TYPE_CHECKING, List

import peewee as pw

from ..logger import logger
from ..db import AutoIncrementField
from . import BaseModel

if TYPE_CHECKING:  # pragma: no cover
    from .job import Job


class Folder(BaseModel):
    if TYPE_CHECKING:  # pragma: no cover
        folder_id: int
        parent: "Folder"
        children: List["Folder"]
        name: str
        jobs: List[Job]
    else:
        folder_id = AutoIncrementField()
        name = pw.CharField()
        parent = pw.ForeignKeyField("self", null=True, backref="children")
        created_at = pw.DateTimeField(default=datetime.datetime.now)
        updated_at = pw.DateTimeField()

    _ignore_save_assert = False

    class Meta:
        indexes = ((("parent", "name"), True),)

    def save(self, *args: Any, **kwargs: Any) -> None:
        if not self._ignore_save_assert:
            assert (
                self.name not in (".", "..", "")
                and "/" not in self.name
                and not self.name.isdigit()
            ), f"Invalid folder name '{self.name}'"
            assert self.parent is not None, "Need to specify a parent folder"
        self._ignore_save_assert = False

        # can never be its own parent
        if self.parent == self:
            raise pw.IntegrityError("Folder can not be its own parent")

        self.updated_at = datetime.datetime.now()

        return super().save(*args, **kwargs)

    @property
    def path(self) -> str:
        if self.parent is None:
            return "/"
        # this will be slow, could optimize with CTE
        return cast(str, os.path.join(self.parent.path, self.name))

    @classmethod
    def get_root(cls) -> "Folder":
        folder = Folder.get_or_none(Folder.parent.is_null(), name="root")
        if folder is None:
            # bypass assertions
            folder = super(BaseModel, cls).create(name="root", _ignore_save_assert=True)
        return cast(Folder, folder)

    @staticmethod
    def find_by_path(cwd: "Folder", path: str) -> Optional["Folder"]:
        if cwd is None:
            return None
        if path == "/":
            return Folder.get_root()
        if path.startswith("/"):
            return Folder.find_by_path(Folder.get_root(), path[1:])
        if path.endswith("/"):
            path = path[:-1]
        if path == "..":
            return cwd.parent
        if path == "" or path == ".":
            return cwd
        if "/" not in path:
            return cwd.subfolder(path)
        head, tail = path.split(os.sep, 1)
        logger.debug("Resolve path %s in %s: %s, %s", path, cwd.path, head, tail)

        if head == "..":
            return Folder.find_by_path(cwd.parent, tail)
        else:
            return Folder.find_by_path(cast(Folder, cwd.subfolder(head)), tail)

    def __truediv__(self, name: str) -> Optional["Folder"]:
        return self.subfolder(name)

    def add_folder(self, name: str) -> "Folder":
        return Folder.create(name=name, parent=self)

    def subfolder(self, name: str) -> Optional["Folder"]:
        return Folder.get_or_none(Folder.parent == self, Folder.name == name)

    def jobs_recursive(self) -> List["Job"]:
        # @TODO: Optimize, hierarchical expression?
        jobs: List[Job] = list(self.jobs)
        for child in self.children:
            jobs += child.jobs_recursive()
        return jobs



from contextlib import contextmanager
from io import StringIO
from unittest.mock import Mock, ANY, MagicMock

import pytest

from kong import config
from kong.drivers.local_driver import LocalDriver
import kong.drivers
from kong.drivers import DriverMismatch
from kong.model import Job, Folder
import peewee as pw


@pytest.fixture
def job(tree):
    return Job.create(
        batch_job_id=42, folder=tree, command="sleep 1", driver=LocalDriver
    )


def test_create(tree):
    j1 = Job.create(batch_job_id=42, folder=tree, command="sleep 2", driver=LocalDriver)
    assert j1.job_id is not None
    job = Job.get(batch_job_id=42)
    assert job == j1
    assert len(tree.jobs) == 1
    assert tree.jobs[0] == j1
    assert j1.folder == tree
    assert j1.command == "sleep 2"

    with pytest.raises(pw.IntegrityError):
        Job.create(batch_job_id=42, command="sleep 4", folder=tree, driver=LocalDriver)
    assert (
        Job.create(batch_job_id=43, folder=tree, command="sleep 4", driver=LocalDriver)
        is not None
    )

    f2 = tree.subfolder("f2")
    j2 = Job.create(batch_job_id=44, folder=f2, command="sleep 4", driver=LocalDriver)
    assert j2 is not None
    assert j2.command == "sleep 4"
    assert len(f2.jobs) == 1
    assert f2.jobs[0] == j2
    assert j2.folder == f2


def test_properties(tree):
    j1 = Job.create(batch_job_id=42, folder=tree, command="a", driver=LocalDriver)
    j1.data["log_dir"] = "LOGDIR_PATH"
    j1.data["output_dir"] = "OUTPUT_PATH"
    j1.save()
    assert j1.log_dir == "LOGDIR_PATH"
    assert j1.output_dir == "OUTPUT_PATH"


def test_set_driver(state, monkeypatch):
    j1 = Job.create(
        batch_job_id=42, folder=state.cwd, command="sleep 2", driver=LocalDriver
    )

    driver = LocalDriver(state.config)

    class DummyDriver:
        pass

    j1.ensure_driver_instance(driver)
    assert j1.driver_instance == driver

    # forcibly unset driver instance
    j1._driver_instance = None
    j1.ensure_driver_instance(state.config)
    assert isinstance(j1.driver_instance, LocalDriver)

    j1._driver_instance = None

    with pytest.raises(DriverMismatch):
        j1.ensure_driver_instance(DummyDriver())
    assert j1._driver_instance is None

    # bypass interface
    monkeypatch.setattr(
        "kong.drivers.driver_base.DriverBase.__abstractmethods__", set()
    )

    class ValidDriver(kong.drivers.driver_base.DriverBase):
        def __init__(self):
            pass

    with pytest.raises(DriverMismatch):
        j1.ensure_driver_instance(ValidDriver())
    assert j1._driver_instance is None

    j2 = Job.create(
        batch_job_id=42, folder=state.cwd, command="sleep 2", driver=ValidDriver
    )

    valid_driver = ValidDriver()
    j2.ensure_driver_instance(valid_driver)
    assert j2.driver_instance == valid_driver

    j2.ensure_driver_instance(driver)  # nothing happens
    assert not isinstance(j2.driver_instance, LocalDriver)


def test_rm(tree, state):
    root = tree

    j1 = Job.create(folder=tree, command="sleep 1", driver=LocalDriver)

    class PseudoDriver(LocalDriver):
        pass

    pseudo_driver = PseudoDriver(state.config)
    pseudo_driver.cleanup = Mock()
    j1._driver_instance = pseudo_driver

    assert j1.folder == root
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    j1.remove()
    root = Folder.get_root()
    assert len(root.jobs) == 0

    pseudo_driver.cleanup.assert_called_once_with(j1)


def test_resubmit(job):
    driver = Mock()
    driver.resubmit = Mock()
    job._driver_instance = driver

    job.resubmit()
    driver.resubmit.assert_called_once_with(job)


def test_submit(job):
    driver = Mock()
    driver.submit = Mock()
    job._driver_instance = driver

    job.submit()
    driver.submit.assert_called_once_with(job)


def test_wait(job):
    driver = Mock()
    driver.wait = Mock()
    job._driver_instance = driver

    job.wait()
    driver.wait.assert_called_once_with(job, timeout=ANY)


def test_get_status(job):
    driver = Mock()
    driver.sync_status = Mock()
    job._driver_instance = driver

    job.get_status()
    driver.sync_status.assert_called_once_with(job)


def test_stdout_stderr(job):
    stdout = StringIO("TESTVALUE")
    stderr = StringIO("2TESTVALUE2")

    @contextmanager
    def get_stdout(self):
        yield stdout

    @contextmanager
    def get_stderr(self):
        yield stderr

    driver = Mock(stdout=get_stdout, stderr=get_stderr)
    job._driver_instance = driver

    with job.stdout() as f:
        assert f.read() == "TESTVALUE"

    with job.stderr() as f:
        assert f.read() == "2TESTVALUE2"


def test_kill(job):
    driver = Mock()
    driver.kill = Mock()
    job._driver_instance = driver

    job.kill()
    driver.kill.assert_called_once_with(job)


def test_bulk_select(state, monkeypatch):
    jobs = [state.create_job(command="sleep 1") for _ in range(50)]
    ids = [j.job_id for j in jobs]

    # make sure it's actually batched
    with monkeypatch.context() as m:
        execute = Mock(return_value=[])
        m.setattr("peewee.BaseQuery.execute", execute)
        list(Job.bulk_select(Job.job_id, ids, batch_size=10))
        assert execute.call_count == 5

    selected = list(Job.bulk_select(Job.job_id, ids, batch_size=10))
    assert len(selected) == len(jobs)
    for exp, act in zip(jobs, selected):
        assert exp.job_id == act.job_id



from setuptools import setup, find_packages  # type: ignore

dev_requires = ["black"]
tests_require = ["pytest", "coverage", "pytest-cov", "mypy", "flake8", "tox"]
setup(
    name="kong-batch",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="",
    url="http://github.com/paulgessinger/kong",
    author="Paul Gessinger",
    author_email="hello@paulgessinger.com",
    license="MIT",
    install_requires=[
        "click",
        "pyyaml",
        "halo",
        "sh",
        "python-dateutil",
        "peewee",
        "coloredlogs",
        "humanfriendly",
        "psutil",
        "schema",
        "tqdm",
        "notifiers",
        "jinja2",
        "pytimeparse",
    ],
    tests_require=tests_require,
    extras_require={"dev": dev_requires, "test": tests_require, "ipython": ["ipython"]},
    entry_points={"console_scripts": ["kong=kong.cli:main"]},
    packages=find_packages("src"),
    package_dir={"": "src"},
)



from ..logger import logger


class InvalidJobStatus(BaseException):
    pass


class DriverMismatch(BaseException):
    pass


def get_driver(value: str) -> type:
    logger.debug("Attempting loading driver %s", value)
    import importlib

    components = value.split(".")
    module_name = ".".join(components[:-1])
    class_name = components[-1]

    module = importlib.import_module(module_name)
    class_: type = getattr(module, class_name)
    return class_



import os
import shutil
from datetime import timedelta
from typing import Collection, Iterator
from unittest.mock import Mock, ANY, call

import pytest

from kong import util
from kong.config import Config, slurm_schema
from kong.drivers import InvalidJobStatus, get_driver
from kong.drivers.htcondor_driver import (
    HTCondorInterface,
    HTCondorDriver,
    HTCondorAccountingItem,
    ShellHTCondorInterface,
)
from kong.model import Job, Folder
from kong.util import is_executable


@pytest.fixture
def driver(monkeypatch, state):
    # set some config values
    data = state.config.data.copy()
    data["htcondor_driver"] = dict()
    state.config = Config(data)

    monkeypatch.setattr(HTCondorInterface, "__abstractmethods__", set())
    monkeypatch.setattr(ShellHTCondorInterface, "__init__", Mock(return_value=None))

    sif = ShellHTCondorInterface()
    sif.config = state.config.htcondor_driver

    return HTCondorDriver(state.config, sif)


def test_condor_q(driver, monkeypatch, state):
    condor_q_output = """
[
{
  "ClusterId": 5184659,
  "JobStatus": 0,
  "ProcId": 0
}
,
{
  "ClusterId": 5184660,
  "JobStatus": 1,
  "ProcId": 0
}
,
{
  "ClusterId": 5184661,
  "JobStatus": 2,
  "ProcId": 0
}
,
{
  "ClusterId": 5184662,
  "JobStatus": 3,
  "ProcId": 0
}
,
{
  "ClusterId": 5184663,
  "JobStatus": 4,
  "ProcId": 0
}
,
{
  "ClusterId": 5184664,
  "JobStatus": 5,
  "ProcId": 0
}
,
{
  "ClusterId": 5184665,
  "JobStatus": 6,
  "ProcId": 0
}
,
{
  "ClusterId": 5184666,
  "JobStatus": 42,
  "ProcId": 0
}
]
    """.strip()

    with monkeypatch.context() as m:

        mock = Mock(return_value=condor_q_output)
        m.setattr(driver.htcondor, "_condor_q", mock)

        res = list(driver.htcondor.condor_q())

        mock.assert_called_once_with(
            "-attributes", "ClusterId,ProcId,JobStatus", "-json"
        )

        ref = [
            HTCondorAccountingItem(5184659, Job.Status.SUBMITTED, -1),
            HTCondorAccountingItem(5184660, Job.Status.SUBMITTED, -1),
            HTCondorAccountingItem(5184661, Job.Status.RUNNING, -1),
            HTCondorAccountingItem(5184662, Job.Status.FAILED, -1),
            HTCondorAccountingItem(5184663, Job.Status.COMPLETED, -1),
            HTCondorAccountingItem(5184664, Job.Status.FAILED, -1),
            HTCondorAccountingItem(5184665, Job.Status.FAILED, -1),
            HTCondorAccountingItem(5184666, Job.Status.UNKNOWN, -1),
        ]

        assert len(ref) == len(res)
        for a, b in zip(ref, res):
            assert a == b


def test_condor_q_empty(driver, monkeypatch, state):
    mock = Mock(return_value="")
    monkeypatch.setattr(driver.htcondor, "_condor_q", mock)
    res = list(driver.htcondor.condor_q())
    assert res == []


def test_condor_history_empty(driver, monkeypatch, state):
    mock = Mock(return_value="")
    monkeypatch.setattr(driver.htcondor, "_condor_history", mock)
    res = list(driver.htcondor.condor_history())
    assert res == []


def test_condor_history(driver, monkeypatch, state):
    condor_history_output = """
[
{
  "ClusterId": 5184659,
  "JobStatus": 0,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184660,
  "JobStatus": 1,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184661,
  "JobStatus": 2,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184662,
  "JobStatus": 3,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184663,
  "JobStatus": 4,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184664,
  "JobStatus": 5,
  "ProcId": 0,
  "ExitCode": 3
}
,
{
  "ClusterId": 5184665,
  "JobStatus": 6,
  "ProcId": 0,
  "ExitCode": 4
}
,
{
  "ClusterId": 5184666,
  "JobStatus": 42,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184667,
  "JobStatus": 4,
  "ProcId": 0,
  "ExitCode": 0
}
,
{
  "ClusterId": 5184668,
  "JobStatus": 4,
  "ProcId": 0,
  "ExitCode": 1
}
]
    """.strip()

    with monkeypatch.context() as m:
        mock = Mock(return_value=condor_history_output)
        m.setattr(driver.htcondor, "_condor_history", mock)

        res = list(driver.htcondor.condor_history())

        mock.assert_called_once_with(
            ANY,
            "-attributes",
            "ClusterId,ProcId,JobStatus,ExitCode",
            "-json",
            "-limit",
            ANY,
        )

        ref = [
            HTCondorAccountingItem(5184659, Job.Status.SUBMITTED, 0),
            HTCondorAccountingItem(5184660, Job.Status.SUBMITTED, 0),
            HTCondorAccountingItem(5184661, Job.Status.RUNNING, 0),
            HTCondorAccountingItem(5184662, Job.Status.FAILED, 0),
            HTCondorAccountingItem(5184663, Job.Status.COMPLETED, 0),
            HTCondorAccountingItem(5184664, Job.Status.FAILED, 3),
            HTCondorAccountingItem(5184665, Job.Status.FAILED, 4),
            HTCondorAccountingItem(5184666, Job.Status.UNKNOWN, 0),
            HTCondorAccountingItem(5184667, Job.Status.COMPLETED, 0),
            HTCondorAccountingItem(5184668, Job.Status.FAILED, 1),
        ]

        assert len(ref) == len(res)
        for a, b in zip(ref, res):
            assert a == b


def test_condor_submit_rm(driver, monkeypatch, state):
    condor_submit_output = """
Submitting job(s).
1 job(s) submitted to cluster {batch_job_id}.
    """.strip()

    with monkeypatch.context() as m:
        batch_job_id = 5184660
        condor_submit = Mock(
            return_value=condor_submit_output.format(batch_job_id=batch_job_id)
        )
        m.setattr(driver.htcondor, "_condor_submit", condor_submit)
        job = Job()
        job.data["batchfile"] = "somefile.sh"

        jid = driver.htcondor.condor_submit(job)
        condor_submit.assert_called_once_with("somefile.sh")
        assert jid == batch_job_id
        job.batch_job_id = jid

        condor_rm = Mock()
        m.setattr(driver.htcondor, "_condor_rm", condor_rm)
        driver.htcondor.condor_rm(job)
        condor_rm.assert_called_once_with(batch_job_id)


def test_repr():
    htai = HTCondorAccountingItem(1, Job.Status.UNKNOWN, 0)
    assert repr(htai) != ""


def test_create_job(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        memory=1500,
        name="job1",
        universe="amazing",
        walltime=timedelta(hours=5),
    )

    extra = 'requirements = (OpSysAndVer =?= "CentOS7")'
    driver.htcondor_config["submitfile_extra"] = extra

    assert j1.status == Job.Status.CREATED
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    assert j1.batch_job_id is None
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    assert os.path.exists(j1.data["jobscript"])
    assert os.path.exists(j1.data["batchfile"])
    assert is_executable(j1.data["jobscript"])

    j2 = driver.create_job(command="sleep 1", walltime="03:00:00", folder=root)
    assert j2.data["walltime"] == 3 * 60 * 60

    with open(j2.data["jobscript"]) as f:
        jobscript = f.read()
        assert str(j2.job_id) in jobscript
        assert str(j2.cores) in jobscript
        assert j2.command in jobscript
        for v in ["output_dir", "log_dir", "stdout"]:
            assert j2.data[v] in jobscript
    with open(j2.data["batchfile"]) as f:
        batchfile = f.read()
        assert str(j2.cores) in batchfile
        assert str(j2.memory) in batchfile
        for v in ["name", "htcondor_out", "universe", "walltime", "jobscript"]:
            assert str(j2.data[v]) in batchfile

    assert extra in batchfile

    with pytest.raises(ValueError):
        driver.create_job(command="sleep 1", walltime="100:00:00", folder=root)

    with pytest.raises(ValueError):
        driver.create_job(command="sleep 1", walltime=42, folder=root)


def test_submit_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        name="job1",
        walltime=timedelta(hours=5),
    )

    assert j1.status == Job.Status.CREATED

    j1.status = Job.Status.SUBMITTED
    j1.save()

    with pytest.raises(InvalidJobStatus):
        driver.submit(j1)

    j1.status = Job.Status.CREATED
    j1.save()

    batch_job_id = 5_207_375

    with monkeypatch.context() as m:
        condor_submit = Mock(return_value=batch_job_id)
        m.setattr(driver.htcondor, "condor_submit", condor_submit)
        driver.submit(j1)
        condor_submit.assert_called_once_with(j1)

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)


def test_resubmit_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    condor_submit = Mock(return_value=batch_job_id)
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.submit(j1)
    condor_submit.assert_called_once_with(j1)

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    monkeypatch.setattr(driver.htcondor, "condor_q", Mock(return_value=[]))
    monkeypatch.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))
    with pytest.raises(InvalidJobStatus):
        driver.resubmit(j1)

    HTAI = HTCondorAccountingItem
    monkeypatch.setattr(
        driver.htcondor,
        "condor_q",
        Mock(return_value=[HTAI(j1.batch_job_id, Job.Status.FAILED, 0)]),
    )

    bjid2 = 42
    condor_submit = Mock(return_value=bjid2)
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)

    with monkeypatch.context() as m:
        # job errors on kill, resubmits anyway
        m.setattr(driver, "kill", Mock(side_effect=RuntimeError()))
        m.setattr("os.path.exists", Mock(side_effect=[True, False, False]))
        m.setattr("os.remove", Mock())
        j1 = driver.resubmit(j1)

    condor_submit.assert_called_once()
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(bjid2)  # gets new batch job id

    with monkeypatch.context() as m:
        m.setattr(driver, "sync_status", Mock())  # disable sync for a second
        with pytest.raises(InvalidJobStatus):
            driver.resubmit(j1)  # stays in SUBMITTED, not accepted

    monkeypatch.setattr(
        driver.htcondor,
        "condor_history",
        Mock(return_value=[HTAI(j1.batch_job_id, Job.Status.FAILED, 0)]),
    )

    # will go to failed

    bjid3 = 99
    condor_submit = Mock(return_value=bjid3)
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    j1 = driver.resubmit(j1)
    condor_submit.assert_called_once()
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(bjid3)


def test_job_bulk_resubmit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
    ]

    other_job = driver.create_job(
        command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
    )
    other_job.status = Job.Status.COMPLETED
    other_job.save()

    jobs[0].status = Job.Status.FAILED
    jobs[0].save()

    condor_submit = Mock(side_effect=[1, 2, 3])
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs[1:])
    assert condor_submit.call_count == 2

    for job in jobs[1:]:
        job.status = Job.Status.COMPLETED

        with open(job.data["stdout"], "w") as f:
            f.write("hurz")

        job.save()

    shutil.rmtree(jobs[0].data["output_dir"])

    # we need to prevent driver from actually calling submit
    submit = Mock()
    remove = Mock(wraps=os.remove)
    makedirs = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver, "submit", submit)
        m.setattr(driver.htcondor, "condor_q", Mock(return_value=[]))
        m.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))
        m.setattr(driver, "bulk_kill", Mock(side_effect=RuntimeError))
        m.setattr("os.remove", remove)
        m.setattr("os.makedirs", makedirs)
        driver.bulk_resubmit(jobs)
    assert submit.call_count == len(jobs)
    remove.assert_has_calls([call(j.data["stdout"]) for j in jobs[1:]], any_order=True)
    makedirs.assert_has_calls(
        [call(j.data["output_dir"]) for j in jobs[1:]], any_order=True
    )

    for job in jobs:
        job.reload()
        assert job.status == Job.Status.CREATED

    # bug: all jobs where reset to created. Check this is not the case anymore
    other_job.reload()
    assert other_job.status != Job.Status.CREATED


def test_resubmit_bulk_invalid_status(driver, state, monkeypatch):
    monkeypatch.setattr(driver, "sync_status", Mock())
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    monkeypatch.setattr(driver, "bulk_sync_status", Mock(return_value=[j1]))
    for status in (Job.Status.CREATED, Job.Status.SUBMITTED, Job.Status.RUNNING):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            driver.bulk_resubmit([j1])


def test_job_bulk_resubmit_no_submit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
    ]

    condor_submit = Mock(side_effect=[1, 2, 3])
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs)
    assert condor_submit.call_count == 3

    for job in jobs:
        job.status = Job.Status.COMPLETED
        job.save()

    bulk_submit = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver.htcondor, "condor_q", Mock(return_value=[]))
        m.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))
        m.setattr(driver, "bulk_submit", bulk_submit)
        driver.bulk_resubmit(jobs, do_submit=False)
    assert bulk_submit.call_count == 0


def test_stdout_stderr(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        name="job1",
        walltime=timedelta(hours=5),
    )

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    condor_submit = Mock(return_value=batch_job_id)
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.submit(j1)

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    stdout = "VALUE VALUE VALUE"

    with open(j1.data["stdout"], "w") as fh:
        fh.write(stdout)

    with driver.stdout(j1) as fh:
        assert stdout == fh.read()

    with pytest.raises(NotImplementedError):
        driver.stderr(j1)


def test_sync_status(driver, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    # with monkeypatch.context() as m:
    #     condor_submit = Mock(return_value=1)
    #     m.setattr(driver.htcondor, "condor_submit", condor_submit)

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    monkeypatch.setattr(
        driver.htcondor, "condor_submit", Mock(return_value=batch_job_id)
    )
    driver.submit(j1)
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    condor_q_return = [
        [HTCondorAccountingItem(batch_job_id, Job.Status.RUNNING, 0)],
        [HTCondorAccountingItem(batch_job_id, Job.Status.FAILED, 0)],
    ]
    condor_q = Mock(side_effect=condor_q_return)
    monkeypatch.setattr(driver.htcondor, "condor_q", condor_q)
    monkeypatch.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))

    j1 = driver.sync_status(j1)
    assert j1.status == Job.Status.RUNNING
    j1 = driver.sync_status(j1)
    assert j1.status == Job.Status.FAILED


def test_bulk_create(driver, state):
    root = Folder.get_root()
    jobs = driver.bulk_create_jobs(
        [{"folder": root, "command": "sleep 1"} for i in range(10)]
    )
    assert len(jobs) == 10
    for job in jobs:
        assert job.status == Job.Status.CREATED


def test_bulk_submit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    assert len(jobs) == 15
    for job in jobs:
        assert job.status == Job.Status.CREATED

    condor_submit = Mock(side_effect=[i + 1 for i in range(len(jobs))])

    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs)

    assert condor_submit.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED
        assert str(job.job_id) == job.batch_job_id


def test_bulk_sync_status(driver, state, monkeypatch):

    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    assert len(jobs) == 15
    for job in jobs:
        assert job.status == Job.Status.CREATED

    condor_submit = Mock(side_effect=[i + 1 for i in range(len(jobs))])
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs)

    HTAI = HTCondorAccountingItem

    condor_q = Mock(
        return_value=[HTAI(i + 1, Job.Status.RUNNING, -1) for i in range(len(jobs))]
    )

    # pretend they're all running now

    with monkeypatch.context() as m:
        m.setattr(driver.htcondor, "condor_q", condor_q)
        m.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))

        jobs = driver.bulk_sync_status(jobs)

        driver.htcondor.condor_history.assert_called_once_with()
        condor_q.assert_called_once_with()

    for job in jobs:
        assert job.status == Job.Status.RUNNING

    with monkeypatch.context() as m:
        condor_history = Mock(
            return_value=[
                HTAI(
                    i + 1,
                    Job.Status.COMPLETED if i < 6 else Job.Status.FAILED,
                    0 if i < 6 else 1,
                )
                for i in range(len(jobs))
            ]
        )
        m.setattr(driver.htcondor, "condor_history", condor_history)
        m.setattr(driver.htcondor, "condor_q", Mock(return_value=[]))

        jobs = driver.bulk_sync_status(jobs)
        condor_history.assert_called_once_with()
        driver.htcondor.condor_q.assert_called_once_with()

    for job in jobs[:6]:
        assert job.status == Job.Status.COMPLETED
    for job in jobs[6:]:
        assert job.status == Job.Status.FAILED


def test_bulk_sync_status_invalid_id(driver, state, monkeypatch):

    root = Folder.get_root()

    jobs = driver.bulk_create_jobs(
        [{"folder": root, "command": "sleep 1"} for i in range(10)]
    )

    condor_submit = Mock(side_effect=[i + 1 for i in range(len(jobs))])
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs)

    HTAI = HTCondorAccountingItem
    condor_q_return = [HTAI(i + 1, Job.Status.RUNNING, 0) for i in range(len(jobs))]
    condor_history_return = [HTAI(12_345_665, Job.Status.UNKNOWN, 0)]
    # pretend they're all running now
    monkeypatch.setattr(driver.htcondor, "condor_q", Mock(return_value=condor_q_return))
    monkeypatch.setattr(
        driver.htcondor, "condor_history", Mock(return_value=condor_history_return)
    )
    jobs = driver.bulk_sync_status(jobs)

    for job in jobs:
        assert job.status == Job.Status.RUNNING


def test_kill_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    driver.kill(j1)
    assert j1.status == Job.Status.FAILED

    j1.status = Job.Status.CREATED

    monkeypatch.setattr(driver.htcondor, "condor_submit", Mock(return_value=1))
    driver.submit(j1)

    assert j1.status == Job.Status.SUBMITTED

    condor_rm = Mock()
    monkeypatch.setattr(driver.htcondor, "condor_rm", condor_rm)
    driver.kill(j1)
    condor_rm.assert_called_once_with(j1)

    assert j1.status == Job.Status.FAILED

    driver.kill(j1)
    assert j1.status == Job.Status.FAILED


def test_bulk_kill(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    for job in jobs:
        assert job.status == Job.Status.CREATED

    condor_submit = Mock(side_effect=[i for i in range(len(jobs))])
    monkeypatch.setattr(driver.htcondor, "condor_submit", condor_submit)
    driver.bulk_submit(jobs)
    assert condor_submit.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED

    condor_rm = Mock()
    monkeypatch.setattr(driver.htcondor, "condor_rm", condor_rm)
    monkeypatch.setattr(driver.htcondor, "condor_q", Mock(return_value=[]))
    monkeypatch.setattr(driver.htcondor, "condor_history", Mock(return_value=[]))

    jobs = driver.bulk_kill(jobs)

    assert condor_rm.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.FAILED


def test_wait(driver, state, monkeypatch):
    root = Folder.get_root()

    class HTCondorInterfaceDummy(HTCondorInterface):
        def __init__(self):
            self.max_job_id = 1
            self.id_map = {}
            self.state_idx = 0
            self.jobs = []

        def condor_q(self) -> Iterator[HTCondorAccountingItem]:
            values = [
                [
                    HTCondorAccountingItem(j.batch_job_id, Job.Status.RUNNING, 0)
                    for j in self.jobs
                ],
                [  # first half done
                    HTCondorAccountingItem(j.batch_job_id, Job.Status.COMPLETED, 0)
                    for j in self.jobs[len(self.jobs) // 2 :]
                ],
                [  # all done
                    HTCondorAccountingItem(j.batch_job_id, Job.Status.COMPLETED, 0)
                    for j in self.jobs
                ],
            ]

            v = values[self.state_idx]
            self.state_idx += 1
            return v

        def condor_history(self) -> Iterator[HTCondorAccountingItem]:
            return []

        def condor_submit(self, job: Job) -> int:
            self.jobs.append(job)
            self.max_job_id += 1
            job.batch_job_id = self.max_job_id
            self.id_map[job] = self.max_job_id
            return self.max_job_id

        def scancel(self, job: Job) -> None:
            pass

    hti = HTCondorInterfaceDummy()
    monkeypatch.setattr(driver, "htcondor", hti)

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    for job in jobs:
        assert job.status == Job.Status.CREATED

    driver.bulk_submit(jobs)

    for job in jobs:
        job.reload()
        i = hti.id_map[job]
        assert job.batch_job_id == str(i)
        assert job.status == Job.Status.SUBMITTED

    with monkeypatch.context() as m:
        condor_q = Mock(wraps=hti.condor_q)
        m.setattr(hti, "condor_q", condor_q)
        driver.wait(jobs, poll_interval=0.01)
        assert condor_q.call_count == 3

    for job in jobs:
        job.reload()
        assert job.status == Job.Status.COMPLETED


def test_wait_single(driver, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB'")

    monkeypatch.setattr(driver, "bulk_sync_status", Mock(return_value=[j1]))

    with pytest.raises(ValueError):
        driver.wait(j1)  # job is in CREATED

    j1.status = Job.Status.COMPLETED
    driver.wait(j1)

    # timeout
    j1.status = Job.Status.RUNNING
    with pytest.raises(TimeoutError):
        driver.wait(j1, timeout=0.05, poll_interval=0.1)

    with pytest.raises(TypeError):
        driver.wait("nope")


def test_cleanup_driver(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))

    j1.status = Job.Status.SUBMITTED
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    j1.status = Job.Status.RUNNING
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    j1.status = Job.Status.COMPLETED

    driver.cleanup(j1)
    assert not os.path.exists(j1.data["log_dir"])
    assert not os.path.exists(j1.data["output_dir"])


def test_cleanup_driver_already_deleted(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    shutil.rmtree(j1.data["log_dir"])

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))

    j1.status = Job.Status.COMPLETED

    assert not os.path.exists(j1.data["log_dir"])
    rmtree = Mock(wraps=util.rmtree)
    with monkeypatch.context() as m:
        m.setattr("kong.drivers.batch_driver_base.rmtree", rmtree)
        driver.cleanup(j1)
    rmtree.assert_has_calls([call(j1.data["output_dir"])])

    assert not os.path.exists(j1.data["log_dir"])
    assert not os.path.exists(j1.data["output_dir"])


def test_job_bulk_cleanup(driver, state, monkeypatch):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    jobs[0].status = Job.Status.RUNNING
    jobs[0].save()

    monkeypatch.setattr(driver, "bulk_sync_status", Mock(side_effect=lambda j: j))

    with pytest.raises(InvalidJobStatus):
        driver.bulk_cleanup(jobs)

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    shutil.rmtree(jobs[0].data["log_dir"])
    jobs[0].status = Job.Status.CREATED
    jobs[0].save()

    rmtree = Mock(side_effect=OSError)
    with monkeypatch.context() as m:
        m.setattr("kong.drivers.batch_driver_base.rmtree", rmtree)
        driver.bulk_cleanup(jobs)
    rmtree.assert_has_calls(
        [
            call(jobs[0].data["output_dir"]),
            call(jobs[1].data["log_dir"]),
            call(jobs[1].data["output_dir"]),
            call(jobs[2].data["log_dir"]),
            call(jobs[2].data["output_dir"]),
        ]
    )

    driver.bulk_cleanup(jobs)

    for job in jobs:
        assert not os.path.exists(job.data["log_dir"])
        assert not os.path.exists(job.data["output_dir"])


def test_remove_driver(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))
    cleanup = Mock(side_effect=lambda j: j)
    monkeypatch.setattr(driver, "cleanup", cleanup)

    j1.status = Job.Status.COMPLETED

    driver.remove(j1)

    cleanup.assert_called_once_with(j1)

    assert Job.get_or_none(job_id=j1.job_id) is None


def test_job_bulk_remove(driver, state, monkeypatch):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]
    for job in jobs:
        assert os.path.exists(job.data["log_dir"]), "Does not create job directory"
        assert os.path.exists(
            job.data["output_dir"]
        ), "Does not create output directory"

    monkeypatch.setattr(driver, "bulk_sync_status", Mock(side_effect=lambda j: j))
    driver.bulk_remove(jobs)

    for job in jobs:
        assert not os.path.exists(
            job.data["log_dir"]
        ), "Driver does not cleanup job directory"
        assert not os.path.exists(
            job.data["output_dir"]
        ), "Driver does not cleanup output directory"


def test_get_htcondor_driver():
    driver_class = get_driver("kong.drivers.htcondor_driver.HTCondorDriver")
    assert driver_class == HTCondorDriver



import typing

class And:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any): ...

class Use:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any): ...

class Optional:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any): ...

class Schema:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        ...

    def validate(self, arg: typing.Any) -> typing.Any:
        ...

    def is_valid(self, arg: typing.Any) -> bool:
        ...

